import logging
from far2l.plugin import PluginBase


log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python Character Map"
    openFrom = ["PLUGINSMENU", "COMMANDLINE", "EDITOR", "VIEWER"]

    def Rebuild(self, hDlg):
        self.info.SendDlgMessage(hDlg, self.ffic.DM_ENABLEREDRAW, 0, 0)
        prefix = ["", "&"]
        for i in range(len(self.symbols)):
            row = i // self.max_col
            col = i % self.max_col
            p = prefix[row == self.cur_row and col == self.cur_col]
            offset = self.first_text_item + row * self.max_col + col
            ch = self.s2f(p + self.symbols[i])
            self.info.SendDlgMessage(
                hDlg, self.ffic.DM_SETTEXTPTR, offset, self.ffi.cast("LONG_PTR", ch)
            )
        self.info.SendDlgMessage(hDlg, self.ffic.DM_ENABLEREDRAW, 1, 0)

    def OpenPlugin(self, OpenFrom):
        symbols = []
        for i in range(256):
            symbols.append(chr(i))
        symbols.extend(
            [
                "Ђ",
                "Ѓ",
                "‚",
                "ѓ",
                "„",
                "…",
                "†",
                "‡",
                "€",
                "‰",
                "Љ",
                "‹",
                "Њ",
                "Ќ",
                "Ћ",
                "Џ",
                "ђ",
                "‘",
                "’",
                "“",
                "”",
                "•",
                "–",
                "—",
                "",
                "™",
                "љ",
                "›",
                "њ",
                "ќ",
                "ћ",
                "џ",
                " ",
                "Ў",
                "ў",
                "Ј",
                "¤",
                "Ґ",
                "¦",
                "§",
                "Ё",
                "©",
                "Є",
                "«",
                "¬",
                "­",
                "®",
                "Ї",
                "°",
                "±",
                "І",
                "і",
                "ґ",
                "µ",
                "¶",
                "·",
                "ё",
                "№",
                "є",
                "»",
                "ј",
                "Ѕ",
                "ѕ",
                "ї",
                "А",
                "Б",
                "В",
                "Г",
                "Д",
                "Е",
                "Ж",
                "З",
                "И",
                "Й",
                "К",
                "Л",
                "М",
                "Н",
                "О",
                "П",
                "Р",
                "С",
                "Т",
                "У",
                "Ф",
                "Х",
                "Ц",
                "Ч",
                "Ш",
                "Щ",
                "Ъ",
                "Ы",
                "Ь",
                "Э",
                "Ю",
                "Я",
                "а",
                "б",
                "в",
                "г",
                "д",
                "е",
                "ж",
                "з",
                "и",
                "й",
                "к",
                "л",
                "м",
                "н",
                "о",
                "п",
                "р",
                "с",
                "т",
                "у",
                "ф",
                "х",
                "ц",
                "ч",
                "ш",
                "щ",
                "ъ",
                "ы",
                "ь",
                "э",
                "ю",
                "я",
                "░",
                "▒",
                "▓",
                "│",
                "┤",
                "╡",
                "╢",
                "╖",
                "╕",
                "╣",
                "║",
                "╗",
                "╝",
                "╜",
                "╛",
                "┐",
                "└",
                "┴",
                "┬",
                "├",
                "─",
                "┼",
                "╞",
                "╟",
                "╚",
                "╔",
                "╩",
                "╦",
                "╠",
                "═",
                "╬",
                "╧",
                "╨",
                "╤",
                "╥",
                "╙",
                "╘",
                "╒",
                "╓",
                "╫",
                "╪",
                "┘",
                "┌",
                "█",
                "▄",
                "▌",
                "▐",
                "▀",
                "∙",
                "√",
                "■",
                "⌠",
                "≈",
                "≤",
                "≥",
                "⌡",
                "²",
                "÷",
                "ą",
                "ć",
                "ę",
                "ł",
                "ń",
                "ó",
                "ś",
                "ż",
                "ź",
                "Ą",
                "Ć",
                "Ę",
                "Ł",
                "Ń",
                "Ó",
                "Ś",
                "Ż",
                "Ź",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
                " ",
            ]
        )
        Items = [
            (
                self.ffic.DI_DOUBLEBOX,
                3,
                1,
                38,
                18,
                0,
                {"Selected": 0},
                0,
                0,
                self.s2f("Character Map"),
                0,
            ),
            (
                self.ffic.DI_BUTTON,
                7,
                17,
                12,
                18,
                0,
                {"Selected": 0},
                1,
                self.ffic.DIF_DEFAULT + self.ffic.DIF_CENTERGROUP,
                self.s2f("OK"),
                0,
            ),
            (
                self.ffic.DI_BUTTON,
                13,
                17,
                38,
                18,
                0,
                {"Selected": 0},
                0,
                self.ffic.DIF_CENTERGROUP,
                self.s2f("Cancel"),
                0,
            ),
            (
                self.ffic.DI_USERCONTROL,
                3,
                2,
                38,
                16,
                0,
                {"Selected": 0},
                0,
                self.ffic.DIF_FOCUS,
                self.ffi.NULL,
                0,
            ),
        ]
        self.cur_row = 0
        self.cur_col = 0
        self.max_col = 32
        self.max_row = len(symbols) // self.max_col
        self.first_text_item = len(Items)
        self.symbols = symbols
        self.text = None

        for i in range(len(symbols)):
            row = i // self.max_col
            col = i % self.max_col
            Items.append(
                (
                    self.ffic.DI_TEXT,
                    5 + col,
                    self.first_text_item - 2 + row,
                    5 + col,
                    self.first_text_item - 2 + row,
                    0,
                    {"Selected": 0},
                    0,
                    0,
                    self.ffi.NULL,
                    0,
                )
            )

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                self.Rebuild(hDlg)
                return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)
            elif Msg == self.ffic.DN_BTNCLICK:
                log.debug(
                    "btn DialogProc({0}, {1})".format(
                        Param1, Param2
                    )
                )
                return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)
            elif Msg == self.ffic.DN_KEY and Param1 == self.first_text_item - 1:
                log.debug(
                    "key DialogProc({0}, {1}, {2}, {3})".format(
                        hDlg, "DN_KEY", Param1, Param2
                    )
                )
                if Param2 == self.ffic.KEY_LEFT:
                    self.cur_col -= 1
                elif Param2 == self.ffic.KEY_UP:
                    self.cur_row -= 1
                elif Param2 == self.ffic.KEY_RIGHT:
                    self.cur_col += 1
                elif Param2 == self.ffic.KEY_DOWN:
                    self.cur_row += 1
                elif Param2 == self.ffic.KEY_ENTER:
                    offset = self.cur_row * self.max_col + self.cur_col
                    self.text = self.symbols[offset]
                    log.debug(
                        "enter:{0} row:{1} col:{2}, ch:{3}".format(
                            offset, self.cur_row, self.cur_col, self.text
                        )
                    )
                    return 0
                elif Param2 == self.ffic.KEY_ESC:
                    return 0
                else:
                    return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)
                if self.cur_col == self.max_col:
                    self.cur_col = 0
                elif self.cur_col == -1:
                    self.cur_col = self.max_col - 1
                if self.cur_row == self.max_row:
                    self.cur_row = 0
                elif self.cur_row == -1:
                    self.cur_row = self.max_row - 1
                self.Rebuild(hDlg)
                return 1
            elif Msg == self.ffic.DN_MOUSECLICK:
                log.debug(
                    "mou DialogProc({0}, {1}, {2}, {3})".format(
                        hDlg, "DN_MOUSECLICK", Param1, Param2
                    )
                )
                ch = Param1 - self.first_text_item
                if ch >= 0:
                    focus = self.info.SendDlgMessage(hDlg, self.ffic.DM_GETFOCUS, 0, 0)
                    log.debug("ch:{0} focus:{1}".format(ch, focus))
                    if focus != self.first_text_item - 1:
                        self.info.SendDlgMessage(
                            hDlg, self.ffic.DM_SETFOCUS, self.first_text_item - 1, 0
                        )
                    self.cur_row = ch // self.max_col
                    self.cur_col = ch % self.max_col
                    self.cur_col = min(max(0, self.cur_col), self.max_col - 1)
                    self.cur_row = min(max(0, self.cur_row), self.max_row - 1)
                    offset = self.cur_row * self.max_col + self.cur_col
                    self.text = self.symbols[offset]
                    self.Rebuild(hDlg)
                    return 0
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        fdi = self.ffi.new("struct FarDialogItem []", Items)
        hDlg = self.info.DialogInit(
            self.info.ModuleNumber,
            -1,
            -1,
            42,
            20,
            self.s2f("Character Map"),
            fdi,
            len(fdi),
            0,
            0,
            DialogProc,
            0,
        )
        res = self.info.DialogRun(hDlg)
        if res == 1 and self.text:
            self.info.FSF.CopyToClipboard(self.s2f(self.text))
        self.info.DialogFree(hDlg)



import logging

from far2lc import CheckForEscape


log = logging.getLogger(__name__)
PROGRESS_WIDTH = 30


class ProgressMessage(object):
    def __init__(self, plugin, title, message, maxvalue):
        super().__init__()
        self.info = plugin.info
        self.ffi = plugin.ffi
        self.ffic = plugin.ffic

        self.visible = False
        self.title = title
        self.message = message
        self.maxvalue = maxvalue
        self.setbar(0)

    def s2f(self, s):
        return self.ffi.new("wchar_t []", s)

    def f2s(self, s):
        return self.ffi.string(self.ffi.cast("wchar_t *", s))

    def adv(self, cmd, par):
        self.info.AdvControl(
            self.info.ModuleNumber,
            cmd,
            self.ffi.cast("void *", par),
            self.ffi.NULL
        )

    def show(self):
        if not self.visible:
            self.adv(self.ffic.ACTL_SETPROGRESSSTATE, self.ffic.PGS_INDETERMINATE)
            self.visible = True
        msg = [ self.s2f(self.title), self.s2f(self.message), self.s2f(self.bar) ]
        cmsg = self.ffi.new("wchar_t *[]", msg)
        self.info.Message(
            self.info.ModuleNumber,
            0,
            self.ffi.NULL,
            cmsg,
            len(msg),
            0
        )

    def hide(self):
        if not self.visible:
            return
        self.adv(self.ffic.ACTL_PROGRESSNOTIFY, 0)
        self.adv(self.ffic.ACTL_SETPROGRESSSTATE, self.ffic.PGS_NOPROGRESS)
        self.visible = False

        PANEL_ACTIVE    =self.ffi.cast("HANDLE", -1)
        PANEL_PASSIVE   =self.ffi.cast("HANDLE", -2)
        self.info.Control(PANEL_ACTIVE, self.ffic.FCTL_REDRAWPANEL, 0, 0)
        self.info.Control(PANEL_PASSIVE, self.ffic.FCTL_REDRAWPANEL, 0, 0)

    def setbar(self, value):
        assert type(value) == int
        assert value <= self.maxvalue

        percent = value * 100 // self.maxvalue

        pv = self.ffi.new('struct PROGRESSVALUE *', (percent, 100))
        self.adv(self.ffic.ACTL_SETPROGRESSVALUE, pv)

        nb = int(percent*PROGRESS_WIDTH // 100)
        self.bar =  '\u2588'*nb + '\u2591'*(PROGRESS_WIDTH-nb)

    def update(self, value):
        self.setbar(value)
        self.show()

    def aborted(self):
        return CheckForEscape()

    def close(self):
        self.hide()



import os
import tempfile
import logging
from far2l.plugin import PluginBase

log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python VTLog"
    openFrom = ["PLUGINSMENU", "EDITOR"]

    def Perform(self):
        eur = self.ffi.new("struct EditorUndoRedo *")
        eur.Command = self.ffic.EUR_BEGIN
        self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)

        if 0:
            # get screen from background task
            count = self.info.FSF.VTEnumBackground(self.ffi.NULL, 0)
            if count > 0:
                log.debug(f'VTEnumBackground={count}')
                con_hnds = self.ffi.new("HANDLE []", count)
                vtssize = self.info.FSF.VTEnumBackground(con_hnds, count)
        else:
            # get screen from background like ctrl+o and F4
            con_hnds = self.ffi.new("HANDLE []", 1)
            con_hnds[0] = self.ffi.NULL

        block = None
        with tempfile.NamedTemporaryFile(delete=False) as fp:
            fp.close()
            #os.unlink(fp.name)
            # what to include in screen dump ?
            #flags = self.ffic.VT_LOGEXPORT_COLORED | self.ffic.VT_LOGEXPORT_WITH_SCREENLINES
            flags = self.ffic.VT_LOGEXPORT_WITH_SCREENLINES
            ok = self.info.FSF.VTLogExport(con_hnds[0], flags, self.s2f(fp.name))
            with open(fp.name, mode='rb') as f:
                block = f.read()
                block=block.replace(b'\x0a', b'\x0d')
            os.unlink(fp.name)

        try:
            block = block.decode('utf-8')
        except:
            log.exception('block.decode')
            return
        s = self.s2f(block)
        self.info.EditorControl(self.ffic.ECTL_INSERTTEXT, s)

        eur = self.ffi.new("struct EditorUndoRedo *")
        eur.Command = self.ffic.EUR_END
        self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)
        log.debug('vtinsert end')

    def OpenPlugin(self, OpenFrom):
        self.Perform()
        return -1



import os
import tempfile
import logging
import sqlite3
from far2l.plugin import PluginVFS
from far2l import fardialogbuilder as fdb

log = logging.getLogger(__name__)


FILE_ATTRIBUTE_DIRECTORY = 0x00000010

PanelModeBrief          = 0
PanelModeMedium         = 1
PanelModeFull           = 2
PanelModeWide           = 3
PanelModeDetailed       = 4
PanelModeDiz            = 5
PanelModeLongDiz        = 6
PanelModeOwners         = 7
PanelModeLinks          = 8
PanelModeAlternative    = 9
PanelModeMax            = 10


class Screen:
    def __init__(self, parent):
        self.parent = parent
        self.hScreen = None

    def __enter__(self):
        self.hScreen = self.parent.info.SaveScreen(0, 0, -1, -1)

    def __exit__(self, *args):
        self.parent.info.RestoreScreen(self.hScreen)


class SqlData:
    def __init__(self, parent, conn):
        self.parent = parent
        self.ffi = parent.ffi
        self.ffic = parent.ffic
        self.s2f = parent.s2f
        self.f2s = parent.f2s

        self.conn = conn
        self.names = []
        self.Items = []
        self.rows = []

    def close(self):
        self.conn = None
        self.parent = None

    def GetCurrentPanelItem(self):
        return self.parent.panel.GetCurrentPanelItem()

class SqlDataHandler(SqlData):
    def __init__(self, parent, conn, tablename):
        super().__init__(parent, conn)
        self.tablename = tablename
        self.offset = 0
        self.limit = 100
        self.getdata()

    def getdata(self):
        stmt = '''\
select
    rowid,
    *
from {}
order by
    rowid
limit {}, {}
'''.format(self.tablename, self.offset, self.limit)
        #log.debug('stmt:{}'.format(stmt))
        cur = self.conn.cursor()
        res = cur.execute(stmt)
        self.desc = res.description
        self.rows = res.fetchall()
        #log.debug('data:{}'.format(self.rows))
        cur.close()
        #log.debug('desc:{}'.format(self.desc))

    def GetOpenPluginInfo(self, OpenInfo):
        #log.debug("usqlite.SqlDataHandler.GetOpenPluginInfo")
        self.paneltitle = self.parent.label+': {}/{}'.format(self.parent.dbname.split('/')[-1], self.tablename)

        ColumnTypes = []
        ColumnWidths = []
        py_pm_py_titles = []
        for n in range(len(self.desc)-1):
            desc = self.desc[n+1]
            py_pm_py_titles.append(self.s2f(desc[0]))
            ColumnTypes.append(f'C{n}')
            ColumnWidths.append('5')
        ColumnWidths[-1] = '0'
        ColumnTypes = ','.join(ColumnTypes)
        ColumnWidths = ','.join(ColumnWidths)
        self.py_pm_titles = self.ffi.new("wchar_t *[]", py_pm_py_titles)
        self.py_ColumnTypes = self.s2f(ColumnTypes)
        self.py_ColumnWidths = self.s2f(ColumnWidths)
        py_pm = [
            self.py_ColumnTypes,  # ColumnTypes
            self.py_ColumnWidths, # ColumnWidths
            self.py_pm_titles,    # ColumnTitles
            0,                    # FullScreen
            0,                    # DetailedStatus
            0,                    # AlignExtensions
            0,                    # CaseConversion
            self.ffi.NULL,        # StatusColumnTypes
            self.ffi.NULL,        # StatusColumnWidths
            [0,0],                # Reserved
        ]
        self.pm = self.ffi.new("struct PanelMode *", py_pm)

        wcn = self.ffi.cast("wchar_t *", self.ffi.NULL)
        py_kbt_normal = [wcn]*12
        py_kbt_ctrl = [wcn]*12
        py_kbt_alt = [wcn]*12
        py_kbt_ctrl_shift = [wcn]*12
        py_kbt_alt_shift = [wcn]*12
        py_kbt_ctrl_alt = [wcn]*12

        py_kbt_normal[4-1]=self.s2f("Edit Row")
        py_kbt_normal[6-1]=self.s2f("+100")
        py_kbt_normal[7-1]=self.s2f("-100")

        self.py_kbt = [
            py_kbt_normal,
            py_kbt_ctrl,
            py_kbt_alt,
            py_kbt_ctrl_shift,
            py_kbt_alt_shift,
            py_kbt_ctrl_alt
        ]
        self.kbt = self.ffi.new("struct KeyBarTitles *", self.py_kbt)

        self.py_curdir = self.s2f(self.parent.dbname.split('/')[-1])
        self.py_paneltitle = self.s2f(self.paneltitle)

        Info = self.ffi.cast("struct OpenPluginInfo *", OpenInfo)
        Info.Flags = (
            self.ffic.OPIF_USEFILTER
            | self.ffic.OPIF_ADDDOTS
            | self.ffic.OPIF_SHOWNAMESONLY
        )
        Info.HostFile = self.ffi.NULL
        Info.CurDir = self.py_curdir
        Info.Format = self.py_paneltitle
        Info.PanelTitle = self.py_paneltitle
        #InfoLines
        #InfoLinesNumber
        Info.PanelModesArray = self.pm
        Info.PanelModesNumber = 1
        Info.StartPanelMode = PanelModeDetailed
        Info.StartSortMode = self.ffic.SM_UNSORTED
        #Info.StartSortOrder
        Info.KeyBar = self.kbt
        #Info.ShortcutData
        #Info.Reserved

    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        #log.debug("usqlite.SqlDataHandler.GetFindData({0}, {1}, {2})".format(PanelItem, ItemsNumber, OpMode))

        self.Items = []
        self.names = []
        if len(self.rows):
            names = []
            items = self.ffi.new("struct PluginPanelItem []", len(self.rows))
            for no in range(len(self.rows)):
                rec = self.rows[no]
                items[no].FindData.nPhysicalSize = rec[0]
                fields = []
                names.append([])
                #log.debug(f'{rec}')
                for fldno in range(1, len(rec)):
                    v = rec[fldno]
                    if isinstance(v, str):
                        fields.append(v)
                    elif isinstance(v, int):
                        fields.append(str(v))
                    else:
                        fields.append(str(v))
                    names[-1].append(self.s2f(fields[-1]))
                names.append(self.ffi.new("wchar_t *[]", names[-1]))
                items[no].CustomColumnData = names[-1]
                items[no].CustomColumnNumber = len(names[-2])
            self.Items = items
            self.names = names

        PanelItem = self.ffi.cast("struct PluginPanelItem **", PanelItem)
        ItemsNumber = self.ffi.cast("int *", ItemsNumber)
        n = len(self.Items)
        p = self.ffi.NULL if n == 0 else self.Items
        PanelItem[0] = p
        ItemsNumber[0] = n
        #log.debug('/usqlite.SqlMetadataHandler.GetFindData')
        return 1

    def FreeFindData(self, PanelItem, ItemsNumber):
        #log.debug("usqlite.SqlDataHandler.FreeFindData({0}, {1}, n.names={2}, n.Items={3})".format(PanelItem, ItemsNumber, len(self.names), len(self.Items)))
        self.Items = []
        self.names = []

    def SetDirectory(self, dirname):
        if dirname == '..':
            self.parent.dbhandler = SqlMetadataHandler(self.parent, self.parent.dbconn)
            return 1
        return 0

    def EditRow(self, rec):
        #log.debug(f'EditRec: {rec}')

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            return self.parent.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        fields = []
        for fldno in range(1, len(rec)):
            name = self.desc[fldno]
            v = rec[fldno]
            if isinstance(v, str):
                pass
            elif isinstance(v, int):
                v = str(v)
            else:
                v = str(v)
            if len(v) > 64:
                v = v[:64]+' ...'
            fields.append(
                fdb.HSizer(
                    fdb.TEXT(None, self.desc[fldno][0]+':'),
                    fdb.TEXT(self.desc[fldno][0], v),
                )
            )
        fields.append(
            fdb.HLine()
        )
        fields.append(
            fdb.HSizer(
                fdb.BUTTON("vcancel", "Cancel", default=True, flags=self.ffic.DIF_CENTERGROUP),
#                fdb.BUTTON("vupdate", "Update", flags=self.ffic.DIF_CENTERGROUP),
            )
        )
        b = fdb.DialogBuilder(
            self.parent,
            DialogProc,
            f"SQLite rec: {self.tablename}/{rec[0]}",
            "helptopic",
            0,
            fdb.VSizer(
                *list(fields)
            ),
        )
        dlg = b.build(-1, -1)
        res = self.parent.info.DialogRun(dlg.hDlg)
        log.debug(f'EditRec: res={res}')
        if res not in (-1, dlg.ID_vcancel):
            #dlg.GetText(dlg.ID_vapath),
            #log.debug(f'saverec: {rec}')
            pass
        self.parent.info.DialogFree(dlg.hDlg)

    def ProcessKey(self, Key, ControlState):
        #log.debug('usqlite.SqlDataHandler.ProcessKey: key:{:x} state:{:x} f4:{}'.format(Key, ControlState, Key == self.ffic.VK_F4))
        if ControlState == 0:
            if Key == self.ffic.VK_F4:
                pnli, pnlidata = self.GetCurrentPanelItem()
                if pnli:
                    rowid = pnli.FindData.nPhysicalSize
                    for rec in self.rows:
                        if rec[0] == rowid:
                            self.EditRow(rec)
                            break
                return True
            elif Key == self.ffic.VK_F6:
                #log.debug('F6')
                self.offset += self.limit
                self.getdata()
                if len(self.rows) == 0:
                    self.offset -= self.limit
                    self.getdata()
                handle = self.ffi.cast("HANDLE", -1)
                self.parent.info.Control(handle, self.ffic.FCTL_UPDATEPANEL, 0, 0)
                self.parent.info.Control(handle, self.ffic.FCTL_REDRAWPANEL, 0, 0)
                return True
            elif Key == self.ffic.VK_F7:
                #log.debug('F7')
                self.offset -= self.limit
                if self.offset < 0:
                    self.offset = 0
                self.getdata()
                handle = self.ffi.cast("HANDLE", -1)
                self.parent.info.Control(handle, self.ffic.FCTL_UPDATEPANEL, 0, 0)
                self.parent.info.Control(handle, self.ffic.FCTL_REDRAWPANEL, 0, 0)
                return True
        return False



class SqlMetadataHandler(SqlData):
    def GetOpenPluginInfo(self, OpenInfo):
        #log.debug("usqlite.SqlMetadataHandler.GetOpenPluginInfo")
        py_pm_py_titles = [
            self.s2f("name"),
            self.s2f("type"),
            self.s2f("count"),
        ]
        self.py_pm_titles = self.ffi.new("wchar_t *[]", py_pm_py_titles)
        py_pm = [
            self.s2f("N,C0,C1"),    # ColumnTypes
            self.s2f("0,5,6"),      # ColumnWidths
            self.py_pm_titles,      # ColumnTitles
            0,                      # FullScreen
            0,                      # DetailedStatus
            0,                      # AlignExtensions
            0,                      # CaseConversion
            self.ffi.NULL,          # StatusColumnTypes
            self.ffi.NULL,          # StatusColumnWidths
            [0,0],                  # Reserved
        ]
        self.pm = self.ffi.new("struct PanelMode *", py_pm)

        wcn = self.ffi.cast("wchar_t *", self.ffi.NULL)
        py_kbt_normal = [wcn]*12
        py_kbt_ctrl = [wcn]*12
        py_kbt_alt = [wcn]*12
        py_kbt_ctrl_shift = [wcn]*12
        py_kbt_alt_shift = [wcn]*12
        py_kbt_ctrl_alt = [wcn]*12

        py_kbt_normal[4-1]=self.s2f("DDL")

        self.py_kbt = [
            py_kbt_normal,
            py_kbt_ctrl,
            py_kbt_alt,
            py_kbt_ctrl_shift,
            py_kbt_alt_shift,
            py_kbt_ctrl_alt
        ]
        self.kbt = self.ffi.new("struct KeyBarTitles *", self.py_kbt)

        self.py_curdir = self.s2f(self.parent.dbname.split('/')[-1])
        self.py_paneltitle = self.s2f(self.parent.label)

        Info = self.ffi.cast("struct OpenPluginInfo *", OpenInfo)
        Info.Flags = (
            self.ffic.OPIF_USEFILTER
            | self.ffic.OPIF_ADDDOTS
            | self.ffic.OPIF_SHOWNAMESONLY
        )
        Info.HostFile = self.ffi.NULL
        Info.CurDir = self.py_curdir
        Info.Format = self.py_paneltitle
        Info.PanelTitle = self.py_paneltitle
        #InfoLines
        #InfoLinesNumber
        Info.PanelModesArray = self.pm
        Info.PanelModesNumber = 1
        Info.StartPanelMode = PanelModeDetailed
        Info.StartSortMode = self.ffic.SM_UNSORTED
        #Info.StartSortOrder
        Info.KeyBar = self.kbt
        #Info.ShortcutData
        #Info.Reserved

    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        #log.debug("usqlite.SqlMetadataHandler.GetFindData({0}, {1}, {2})".format(PanelItem, ItemsNumber, OpMode))

        stmt = '''\
select
    name, type, sql
from sqlite_master
'''
        cur = self.conn.cursor()
        res = cur.execute(stmt)
        rows = res.fetchall()
        cur.close()

        if len(rows):
            names = []
            items = self.ffi.new("struct PluginPanelItem []", len(rows))
            for no in range(len(rows)):
                rec = rows[no]
                names.append(self.s2f(rec[0])) # name
                items[no].FindData.lpwszFileName = names[-1]
                if rec[1] in ('table', 'view'):
                    items[no].FindData.dwFileAttributes = FILE_ATTRIBUTE_DIRECTORY
                    cur = self.conn.cursor()
                    res = cur.execute('select count(*) as cnt from '+rec[0])
                    n = str(res.fetchone()[0])
                    cur.close()
                else:
                    n = '--'
                names.append(n)
                names.append([self.s2f(rec[1]), self.s2f(names[-1])]) # type, size
                names.append(self.ffi.new("wchar_t *[]", names[-1]))
                items[no].CustomColumnData = names[-1]
                items[no].CustomColumnNumber = 2
            self.Items = items
            self.names = names
            self.rows = rows

        PanelItem = self.ffi.cast("struct PluginPanelItem **", PanelItem)
        ItemsNumber = self.ffi.cast("int *", ItemsNumber)
        n = len(self.Items)
        p = self.ffi.NULL if n == 0 else self.Items
        PanelItem[0] = p
        ItemsNumber[0] = n
        #log.debug('/usqlite.SqlMetadataHandler.GetFindData')
        return 1

    def FreeFindData(self, PanelItem, ItemsNumber):
        #log.debug("usqlite.SqlMetadataHandler.FreeFindData({0}, {1}, n.names={2}, n.Items={3})".format(PanelItem, ItemsNumber, len(self.names), len(self.Items)))
        self.Items = []
        self.names = []

    def SetDirectory(self, dirname):
        if dirname == '..':
            self.parent.info.Control(self.parent.hplugin, self.ffic.FCTL_CLOSEPLUGIN, 0, 0)
            return 1
        for rec in self.rows:
            if rec[0] == dirname and rec[1] in ('table', 'view'):
                self.parent.dbhandler = SqlDataHandler(self.parent, self.parent.dbconn, dirname)
                return 1
        return 0

    def ProcessKey(self, Key, ControlState):
        #log.debug('usqlite.SqlMetadataHandler.ProcessKey: key:{:x} state:{:x} f4:{}'.format(Key, ControlState, Key == self.ffic.VK_F4))
        if Key == self.ffic.VK_F4:
            if ControlState == 0:
                pnli, pnlidata = self.GetCurrentPanelItem()
                if pnli:
                    name = self.f2s(pnli.FindData.lpwszFileName)
                    #log.debug('F4: rc={} {}'.format(rc, name))
                    for rec in self.rows:
                        if rec[0] == name:
                            fd, fname = tempfile.mkstemp(suffix='.sql', prefix=None, dir='/tmp', text=True)
                            os.close(fd)
                            with open(fname, 'wt') as fp:
                                fp.write(rec[2])
                            self.parent.info.Editor(
                                fname,
                                fname,
                                0, 0, -1, -1,
                                self.ffic.EF_DISABLEHISTORY,#|self.ffic.EF_DELETEONCLOSE,
                                0,
                                0,
                                0xFFFFFFFF,  # =-1=self.ffic.CP_AUTODETECT
                            )
                            os.unlink(fname)
                            break
                return True
            #elif ControlState & self.ffic.PKF_SHIFT:
        return False



class Plugin(PluginVFS):
    label = "Python usqlite"
    openFrom = ["PLUGINSMENU", "DISKMENU"]

    dbname = ''

    def __init__(self, parent, info, ffi, ffic, name=None, conn=None):
        super().__init__(parent, info, ffi, ffic)
        self.dbname = name
        self.dbconn = conn
        self.dbhandler = SqlMetadataHandler(self, self.dbconn)

    def Close(self):
        if self.dbhandler:
            self.dbhandler.close()
        if self.dbconn:
            self.dbconn.close()
        self.dbconn = None
        self.dbhandler = None
        super().Close()

    @staticmethod
    def OpenFilePlugin(parent, info, ffi, ffic, Name, Data, DataSize, OpMode):
        data = b''.join(ffi.cast("char *", Data)[0:min(DataSize, 16)])
        if data != b'SQLite format 3\0':
            return False
        name = ffi.string(ffi.cast("wchar_t *", Name))
        try:
            conn = sqlite3.connect(name)
        except:
            return False
        #log.debug('{} - {}'.format(Plugin.label, name))
        self = Plugin(parent, info, ffi, ffic, name, conn)
        return self

    def GetOpenPluginInfo(self, OpenInfo):
        return self.dbhandler.GetOpenPluginInfo(OpenInfo)

    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        return self.dbhandler.GetFindData(PanelItem, ItemsNumber, OpMode)

    def FreeFindData(self, PanelItem, ItemsNumber):
        return self.dbhandler.FreeFindData(PanelItem, ItemsNumber)

    def SetDirectory(self, Dir, OpMode):
        if OpMode & self.ffic.OPM_FIND:
            return 0
        dirname = self.f2s(Dir)
        #log.debug('usqlite.SetDirectory({}, {})'.format(dirname, self.dbhandler))
        return self.dbhandler.SetDirectory(dirname)

    def ProcessKey(self, Key, ControlState):
        #log.debug("ProcessKey({0}, {1})".format(Key, ControlState))
        return self.dbhandler.ProcessKey(Key, ControlState)




"""
First of all this plugin is an example of how easy writing plugins
for FAR should be when having proper pythonic library.

yfar is a work-in-progress attempt to write such library.

This plugin allows to jump between selected files on the panel.
This was very helpful if you have thousands or even tens of thousands 
files in the directory, select some of them by mask and want to figure 
out which files were actually selected. For convenience bind
them to hotkeys likes Alt+Up / Alt+Down.
"""

__author__ = 'Yaroslav Yanovsky'

from yfar import FarPlugin


class Plugin(FarPlugin):
    label = 'Jump Between Selected Files'
    openFrom = ['PLUGINSMENU', 'FILEPANEL']

    def OpenPlugin(self, _):
        panel = self.get_panel()
        option = self.menu(('Jump to &Previous Selected File',
                            'Jump to &Next Selected File'), self.label)
        if option == 0:  # move up
            for f in reversed(panel.selected):
                if f.index < panel.cursor:
                    panel.cursor = f.index
                    break
        elif option == 1:  # move down
            for f in panel.selected:
                if f.index > panel.cursor:
                    panel.cursor = f.index
                    break
                


import logging
from far2l.plugin import PluginBase
import far2l.fardialogbuilder as dlgb


log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python ED Sort"
    openFrom = ["PLUGINSMENU", "EDITOR"]

    def Dialog(self):
        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                dlg.SetText(dlg.ID_vcol, "")
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        b = dlgb.DialogBuilder(
            self,
            DialogProc,
            "Python uedsort",
            "helptopic",
            0,
            dlgb.VSizer(
                dlgb.HSizer(
                    dlgb.TEXT(None, "Column:"),
                    dlgb.Spacer(),
                    dlgb.EDIT("vcol", 4, maxlength=4),
                ),
                dlgb.HSizer(
                    dlgb.BUTTON("vok", "Ok", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    dlgb.BUTTON("vclose", "Close", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        if res == dlg.ID_vok:
            vcol = int(dlg.GetText(dlg.ID_vcol) or "0", 10)
        else:
            vcol = None
        self.info.DialogFree(dlg.hDlg)
        return vcol

    def Perform(self):
        vcol = self.Dialog()
        if vcol is None:
            return

        ei = self.ffi.new("struct EditorInfo *")
        self.info.EditorControl(self.ffic.ECTL_GETINFO, ei)

        if ei.BlockType != self.ffic.BTYPE_STREAM:
            # working only if not vertical block selection
            return

        egs = self.ffi.new("struct EditorGetString *")
        egs.StringNumber = ei.BlockStartLine
        self.info.EditorControl(self.ffic.ECTL_GETSTRING, egs)
        if egs.SelEnd != -1:
            return

        eur = self.ffi.new("struct EditorUndoRedo *")
        eur.Command = self.ffic.EUR_BEGIN
        self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)

        lines = []
        for lno in range(ei.BlockStartLine, ei.TotalLines):
            esp = self.ffi.new("struct EditorSetPosition *")
            esp.CurLine = lno
            esp.CurPos = esp.Overtype = 0
            esp.CurTabPos = esp.TopScreenLine = esp.LeftPos = -1
            self.info.EditorControl(self.ffic.ECTL_SETPOSITION, esp)

            egs = self.ffi.new("struct EditorGetString *")
            egs.StringNumber = -1
            self.info.EditorControl(self.ffic.ECTL_GETSTRING, egs)

            if egs.SelStart == -1 or egs.SelStart == egs.SelEnd:
                # Stop when reaching the end of the text selection
                break

            if egs.SelEnd != -1:
                # Extend selection to line end
                es = self.ffi.new("struct EditorSelect *")
                es.BlockType = ei.BlockType
                es.BlockStartLine = ei.BlockStartLine
                es.BlockStartPos = 0
                es.BlockWidth = 0
                es.BlockHeight = line - ei.BlockStartLine + 1
                self.info.EditorControl(self.ffic.ECTL_SELECT, es)

            lines.append(self.f2s(egs.StringText))

        self.info.EditorControl(self.ffic.ECTL_DELETEBLOCK, self.ffi.NULL)

        lines.sort(key=lambda x:x[vcol:])
        s = '\r'.join(lines)+'\r'
        s = self.s2f(s)
        self.info.EditorControl(self.ffic.ECTL_INSERTTEXT, s)

        eur = self.ffi.new("struct EditorUndoRedo *")
        eur.Command = self.ffic.EUR_END
        self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)


    def OpenPlugin(self, OpenFrom):
        if OpenFrom == 5:
            # EDITOR
            self.Perform()
        return -1



import os
import os.path
import sys
import types
import configparser
import logging
import logging.config

USERHOME = os.path.expanduser("~/.config/far2l/plugins/python")

logging.basicConfig(level=logging.INFO)


def setup():
    if os.path.isdir(USERHOME):
        sys.path.insert(1, USERHOME)
        fname = os.path.join(USERHOME, "logger.ini")
        if os.path.isfile(fname):
            with open(fname, "rt") as fp:
                ini = configparser.ConfigParser()
                ini.read_file(fp)
                logging.config.fileConfig(ini)


setup()

log = logging.getLogger(__name__)
log.debug("%s start" % ("*" * 20))
log.debug("sys.path={0}".format(sys.path))
log.debug("userhome={0}".format(USERHOME))


def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    log.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


def handle_error(func):
    def __inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            exc_type, exc_value, exc_tb = sys.exc_info()
            handle_exception(exc_type, exc_value, exc_tb)
    return __inner

from .plugin import PluginBase

# commands in shell window:
#     py:load <python modulename>
#     py:unload <registered python module name>

from .far2lcffi import ffi, ffic


class PluginManager:
    Info = None

    def __init__(self):
        self.openplugins = {}
        self.plugins = []

    @handle_error
    def SetStartupInfo(self, Info):
        log.debug("SetStartupInfo({0:08X})".format(Info))
        self.info = ffi.cast("struct PluginStartupInfo *", Info)
        self.ffi = ffi
        self.ffic = ffic
        fname = os.path.join(USERHOME, "plugins.ini")
        if os.path.isfile(fname):
            with open(fname, "rt") as fp:
                ini = configparser.ConfigParser()
                ini.read_file(fp)
                if ini.has_section('autoload'):
                    for name in ini.options('autoload'):
                        self.pluginInstall(name, ini.get('autoload', name)=='open')

    @handle_error
    def pluginRemove(self, name):
        log.debug("remove plugin: {0}".format(name))
        for hplugin, plugin in self.openplugins.items():
            if plugin.name == name:
                del self.openplugins[hplugin]
                break
        for i in range(len(self.plugins)):
            if self.plugins[i].Plugin.name == name:
                del self.plugins[i]
                del sys.modules[name]
                for i in range(len(self.plugins)):
                    self.plugins[i].Plugin.number = i
                return
        log.error("install plugin: {0} - not installed".format(name))

    @handle_error
    def pluginInstall(self, name, autorun=False):
        log.debug(f"install plugin: {name} autorun:{autorun}")
        for i in range(len(self.plugins)):
            if self.plugins[i].Plugin.name == name:
                log.error("install plugin: {0} - allready installed".format(name))
                return
        # plugin = __import__(name, globals(), locals(), [], 0)
        plugin = __import__(name)
        cls = getattr(plugin, "Plugin", None)
        if type(cls) == type(PluginBase) and issubclass(cls, PluginBase):
            # inject plugin name
            cls.USERHOME = USERHOME
            cls.name = name
            self.plugins.append(plugin)
            for i in range(len(self.plugins)):
                self.plugins[i].Plugin.number = i
            if autorun:
                plugin = plugin.Plugin(self, self.info, ffi, ffic)
                hplugin = id(plugin)
                self.openplugins[hplugin] = plugin
        else:
            log.error("install plugin: {0} - not a far2l python plugin".format(name))
            del sys.modules[name]

    def pluginGet(self, hPlugin):
        v = self.openplugins.get(hPlugin, None)
        if v is None:

            class Nop:
                def __getattr__(self, name):
                    log.debug("Nop.__getattr__({})".format(name))
                    return self

                def __call__(self, *args):
                    log.debug("Nop.__call__({0})".format(args))
                    return None

            v = Nop()
            log.error("unhandled pluginGet{0}".format(hPlugin))
        return v

    @handle_error
    def pluginGetFrom(self, OpenFrom, Item):
        id2name = {
            ffic.OPEN_DISKMENU: "DISKMENU",
            ffic.OPEN_PLUGINSMENU: "PLUGINSMENU",
            ffic.OPEN_FINDLIST: "FINDLIST",
            ffic.OPEN_SHORTCUT: "SHORTCUT",
            ffic.OPEN_COMMANDLINE: "COMMANDLINE",
            ffic.OPEN_EDITOR: "EDITOR",
            ffic.OPEN_DIALOG: "DIALOG",
            ffic.OPEN_VIEWER: "VIEWER",
            ffic.OPEN_FILEPANEL: "FILEPANEL",
        }
        name = id2name[OpenFrom]
        log.debug("pluginGetFrom({0} ({1}), {2})".format(OpenFrom, name, Item))
        if OpenFrom in [ffic.OPEN_DISKMENU, ffic.OPEN_FINDLIST]:
            for plugin in self.plugins:
                openFrom = plugin.Plugin.openFrom
                log.debug(
                    "pluginGetFrom(openok={0}, no={1} : {2})".format(
                        name in openFrom, Item, plugin.Plugin.name
                    )
                )
                if name in openFrom:
                    if not Item:
                        return plugin
                    Item -= 1
        elif OpenFrom == ffic.OPEN_DIALOG:
            plugins = []
            for plugin in self.plugins:
                openFrom = plugin.Plugin.openFrom
                log.debug(
                    "pluginGetFrom(openok={0}, no={1} : {2})".format(
                        name in openFrom, Item, plugin.Plugin.name
                    )
                )
                if name in openFrom:
                    plugins.append(plugin)
            if len(plugins) == 1:
                plugin = plugins[0]
            elif len(plugins) > 1:
                names = [p.name for p in plugins]
                n = self.menu(names, 'Select plugin')
                plugin = plugins[n]
            else:
                return None
            return plugin
        elif Item < len(self.plugins):
            return self.plugins[Item]
        return None

    def menu(self, names, title='', selected=0):
        """
        Simple menu. ``names`` is a list of items. Optional
        ``title`` can be provided.
        """
        items = self.ffi.new('struct FarMenuItem []', len(names))
        refs = []
        for i, name in enumerate(names):
            item = items[i]
            item.Checked = item.Separator = 0
            item.Selected = i == selected
            item.Text = txt = self.s2f(name)
            refs.append(txt)
        title = self.s2f(title)
        NULL = self.ffi.NULL
        return self.info.Menu(self.info.ModuleNumber, -1, -1, 0, 
                              self.ffic.FMENU_AUTOHIGHLIGHT
                              | self.ffic.FMENU_WRAPMODE, title,
                              NULL, NULL, NULL, NULL, items, len(items));

    def s2f(self, s):
        return self.ffi.new("wchar_t []", s)

    def f2s(self, s):
        return self.ffi.string(self.ffi.cast("wchar_t *", s))

    def Message(self, lines):
        _MsgItems = [
            self.s2f("Python"),
            self.s2f(""),
        ]
        for line in lines:
            _MsgItems.append(self.s2f(line))
        _MsgItems.extend(
            [
                self.s2f(""),
                self.s2f("\x01"),
                self.s2f("&Ok"),
            ]
        )
        # log.debug('_msgItems: %s', _MsgItems)
        MsgItems = self.ffi.new("wchar_t *[]", _MsgItems)
        self.info.Message(
            self.info.ModuleNumber,  # GUID
            self.ffic.FMSG_WARNING | self.ffic.FMSG_LEFTALIGN,  # Flags
            self.s2f("Contents"),  # HelpTopic
            MsgItems,  # Items
            len(MsgItems),  # ItemsNumber
            1,  # ButtonsNumber
        )

    # manager API
    @handle_error
    def ExitFAR(self):
        log.debug("ExitFAR()")
        for hplugin, plugin in self.openplugins.items():
            plugin.ExitFAR()

    @handle_error
    def GetMinFarVersion(self):
        log.debug("GetMinFarVersion()")

    @handle_error
    def GetPluginInfo(self, Info):
        # log.debug("GetPluginInfo({0:08X})".format(Info))
        Info = self.ffi.cast("struct PluginInfo *", Info)
        self._DiskItems = []
        self._MenuItems = []
        self._ConfigItems = []
        log.debug("GetPluginInfo")
        for plugin in self.plugins:
            openFrom = plugin.Plugin.openFrom
            if "DISKMENU" in openFrom:
                self._DiskItems.append(self.s2f(plugin.Plugin.label))
            if "PLUGINSMENU" in openFrom:
                self._MenuItems.append(self.s2f(plugin.Plugin.label))
            if plugin.Plugin.Configure is not None:
                self._ConfigItems.append(self.s2f(plugin.Plugin.label))
        self.DiskItems = self.ffi.new("wchar_t *[]", self._DiskItems)
        self.MenuItems = self.ffi.new("wchar_t *[]", self._MenuItems)
        self.ConfigItems = self.ffi.new("wchar_t *[]", self._ConfigItems)
        Info.Flags = self.ffic.PF_EDITOR | self.ffic.PF_VIEWER | self.ffic.PF_DIALOG
        Info.DiskMenuStrings = self.DiskItems
        Info.DiskMenuStringsNumber = len(self._DiskItems)
        Info.PluginMenuStrings = self.MenuItems
        Info.PluginMenuStringsNumber = len(self._MenuItems)
        Info.PluginConfigStrings = self.ConfigItems
        Info.PluginConfigStringsNumber = len(self._ConfigItems)
        self._commandprefix = self.s2f("py")
        Info.CommandPrefix = self._commandprefix

    @handle_error
    def OpenPlugin(self, OpenFrom, Item):
        log.debug("OpenPlugin({0}, {1})".format(OpenFrom, Item))
        if OpenFrom == self.ffic.OPEN_COMMANDLINE:
            line = self.f2s(Item)
            log.debug("cmd:{0}".format(line))
            line = line.lstrip()
            linesplit = line.split(" ", 1)
            if linesplit[0] == "unload":
                if len(linesplit) > 1:
                    self.pluginRemove(linesplit[1])
                else:
                    log.debug("missing plugin name in py:unload <plugin name>")
                    self.Message(
                        [
                            "Usage is:",
                            "py:unload <plugin name>",
                        ]
                    )
            elif linesplit[0] == "load":
                if len(linesplit) > 1:
                    line = line.split()
                    self.pluginInstall(line[1], len(line) >2 and line[2] == 'open')
                else:
                    log.debug("missing plugin name in py:load <plugin name> [open]")
                    self.Message(
                        [
                            "Usage is:",
                            "py:load <plugin name> [open]",
                        ]
                    )
            else:
                for plugin in self.plugins:
                    if plugin.Plugin.HandleCommandLine(linesplit[0]) is True:
                        plugin = plugin.Plugin(self, self.info, ffi, ffic)
                        plugin.CommandLine(line)
                        return
                msg = "no plugin handler for command: {}".format(line)
                log.debug(msg)
                self.Message([msg])
            return
        plugin = self.pluginGetFrom(OpenFrom, Item)
        if plugin is not None:
            plugin = plugin.Plugin(self, self.info, ffi, ffic)
            rc = plugin.OpenPlugin(OpenFrom)
            if rc not in (-1, None):
                rc = id(plugin)
                self.openplugins[rc] = plugin
        else:
            rc = None
        return rc

    @handle_error
    def ClosePlugin(self, hPlugin):
        # log.debug("ClosePlugin %08X" % hPlugin)
        plugin = self.openplugins.get(hPlugin, None)
        if plugin is not None:
            plugin.Close()
            del self.openplugins[hPlugin]

    @handle_error
    def OpenFilePlugin(self, Name, Data, DataSize, OpMode):
        log.debug(
            "OpenFilePlugin({0}, {1}, {2}, {3})".format(Name, Data, DataSize, OpMode)
        )
        rc = None
        for plugin in self.plugins:
            plugin = plugin.Plugin.OpenFilePlugin(self, self.info, ffi, ffic, Name, Data, DataSize, OpMode)
            if plugin is not False:
                rc = id(plugin)
                self.openplugins[rc] = plugin
                break
        return rc

    @handle_error
    def ProcessDialogEvent(self, Event, Param):
        for hplugin, plugin in self.openplugins.items():
            rc = plugin.ProcessDialogEvent(Event, Param)
            if rc:
                return rc
        return 0

    @handle_error
    def ProcessEditorEvent(self, Event, Param):
        for hplugin, plugin in self.openplugins.items():
            rc = plugin.ProcessEditorEvent(Event, Param)
            if rc:
                return rc
        return 0

    @handle_error
    def ProcessEditorInput(self, Rec):
        for hplugin, plugin in self.openplugins.items():
            rc = plugin.ProcessEditorInput(Rec)
            if rc:
                return rc
        return 0

    @handle_error
    def ProcessSynchroEvent(self, Event, Param):
        for hplugin, plugin in self.openplugins.items():
            rc = plugin.ProcessSynchroEvent(Event, Param)
            if rc:
                return rc
        return 0

    @handle_error
    def ProcessViewerEvent(self, Event, Param):
        for hplugin, plugin in self.openplugins.items():
            rc = plugin.ProcessViewerEvent(Event, Param)
            if rc:
                return rc
        return 0

    # common plugin functions
    @handle_error
    def GetOpenPluginInfo(self, hPlugin, OpenInfo):
        plugin = self.pluginGet(hPlugin)
        return plugin.GetOpenPluginInfo(OpenInfo)

    @handle_error
    def Configure(self, ItemNumber):
        log.debug("Configure({0})".format(ItemNumber))
        for plugin in self.plugins:
            if plugin.Plugin.Configure is not None:
                if ItemNumber == 0:
                    plugin = plugin.Plugin(self, self.info, ffi, ffic)
                    plugin.Configure()
                    return
                ItemNumber -= 1

    @handle_error
    def ProcessEvent(self, hPlugin, Event, Param):
        # log.debug("ProcessEvent({0}, {1}, {2})".format(hPlugin, Event, Param))
        plugin = self.pluginGet(hPlugin)
        return plugin.ProcessEvent(Event, Param)

    @handle_error
    def ProcessKey(self, hPlugin, Key, ControlState):
        # log.debug("ProcessKey({0}, {1}, {2})".format(hPlugin, Key, ControlState))
        plugin = self.pluginGet(hPlugin)
        return plugin.ProcessKey(Key, ControlState)

    # VFS functions
    @handle_error
    def Compare(self, hPlugin, PanelItem1, PanelItem2, Mode):
        plugin = self.pluginGet(hPlugin)
        return plugin.Compare(PanelItem1, PanelItem2, Mode)

    @handle_error
    def DeleteFiles(self, hPlugin, PanelItem, ItemsNumber, OpMode):
        plugin = self.pluginGet(hPlugin)
        return plugin.DeleteFiles(PanelItem, ItemsNumber, OpMode)

    @handle_error
    def FreeFindData(self, hPlugin, PanelItem, ItemsNumber):
        plugin = self.pluginGet(hPlugin)
        return plugin.FreeFindData(PanelItem, ItemsNumber)

    @handle_error
    def FreeVirtualFindData(self, hPlugin, PanelItem, ItemsNumber):
        plugin = self.pluginGet(hPlugin)
        return plugin.FreeVirtualFindData(PanelItem, ItemsNumber)

    @handle_error
    def GetFiles(self, hPlugin, PanelItem, ItemsNumber, Move, DestPath, OpMode):
        plugin = self.pluginGet(hPlugin)
        return plugin.GetFiles(PanelItem, ItemsNumber, Move, DestPath, OpMode)

    @handle_error
    def GetFindData(self, hPlugin, PanelItem, ItemsNumber, OpMode):
        plugin = self.pluginGet(hPlugin)
        return plugin.GetFindData(PanelItem, ItemsNumber, OpMode)

    @handle_error
    def GetVirtualFindData(self, hPlugin, PanelItem, ItemsNumber, Path):
        plugin = self.pluginGet(hPlugin)
        return plugin.GetVirtualFindData(PanelItem, ItemsNumber, Path)

    @handle_error
    def MakeDirectory(self, hPlugin, Name, OpMode):
        plugin = self.pluginGet(hPlugin)
        return plugin.MakeDirectory(Name, OpMode)

    @handle_error
    def ProcessHostFile(self, hPlugin, PanelItem, ItemsNumber, OpMode):
        plugin = self.pluginGet(hPlugin)
        return plugin.ProcessHostFile(PanelItem, ItemsNumber, OpMode)

    @handle_error
    def PutFiles(self, hPlugin, PanelItem, ItemsNumber, Move, SrcPath, OpMode):
        plugin = self.pluginGet(hPlugin)
        return plugin.PutFiles(PanelItem, ItemsNumber, Move, SrcPath, OpMode)

    @handle_error
    def SetDirectory(self, hPlugin, Dir, OpMode):
        plugin = self.pluginGet(hPlugin)
        return plugin.SetDirectory(Dir, OpMode)

    @handle_error
    def SetFindList(self, hPlugin, PanelItem, ItemsNumber):
        plugin = self.pluginGet(hPlugin)
        return plugin.SetFindList(PanelItem, ItemsNumber)

    @handle_error
    def MayExitFAR(self):
        #log.debug("MayExitFAR(): %s", self.openplugins.items())
        for hplugin, plugin in self.openplugins.items():
            log.debug(plugin.label)
            if not plugin.MayExitFAR():
                return 0
        return 1

    # not exposed from CPython
    @handle_error
    def Analyse(self, pData):
        #pData = AnalyseData
        return 0

    # not exposed from CPython
    @handle_error
    def GetCustomData(self, FilePath, CustomData):
        return 0

    # not exposed from CPython
    @handle_error
    def FreeCustomData(self, CustomData):
        pass



class Panel:
    def __init__(self, plugin):
        self.info = plugin.info
        self.ffic = plugin.ffic
        self.ffi = plugin.ffi
        self.s2f = plugin.s2f
        self.f2s = plugin.f2s

    def _handle(self, active):
        return self.ffi.cast("HANDLE", -1 if active else -2)

    def SetViewMode(self, mode, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_SETVIEWMODE, mode, 0)

    def SetSortMode(self, mode, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_SETSORTMODE, mode, 0)

    def SetNumericSort(self, mode, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_SETNUMERICSORT, mode, 0)

    def SetCaseSensitiveSort(self, mode, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_SETCASESENSITIVESORT, mode, 0)

    def GetPanelPluginHandle(self, active=True):
        result = self.ffi.new("HANDLE *")
        self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELPLUGINHANDLE, 0, self.ffi.cast("LONG_PTR", result))
        return result

    def SetPanelLocation(self, location, active=True):
        # FarPanelLocation *
        return self.info.Control(self._handle(active), self.ffic.FCTL_SETPANELLOCATION, 0, self.ffi.cast("LONG_PTR", result))

    def SetSortOrder(self, ascending, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_SETSORTORDER, 1 if ascending else 0, 0)

    def SetDirectoriesFirst(self, on, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_SETDIRECTORIESFIRST, 1 if on else 0, 0)

    def ClosePlugin(self, arg, active=True):
        # wchar_t *
        return self.info.Control(self._handle(active), self.ffic.FCTL_CLOSEPLUGIN, 0, self.ffi.cast("LONG_PTR", arg))

    def GetPanelInfo(self, active=True):
        pnl = self.ffi.new("struct PanelInfo *")
        if self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELINFO, 0, self.ffi.cast("LONG_PTR", pnl)):
            return pnl
        return None

    def GetPanelHostFile(self, active=True):
        nb = self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELHOSTFILE, 0, 0)
        data = self.ffi.new("wchar_t []", nb)
        self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELHOSTFILE, nb, self.ffi.cast("LONG_PTR", data))
        return self.f2s(data)

    def GetPanelFormat(self, active=True):
        nb = self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELFORMAT, 0, 0)
        data = self.ffi.new("wchar_t []", nb)
        self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELFORMAT, nb, self.ffi.cast("LONG_PTR", data))
        return self.f2s(data)

    def GetPanelDir(self, active=True):
        nb = self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELDIR, 0, 0)
        data = self.ffi.new("wchar_t []", nb)
        self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELDIR, nb, self.ffi.cast("LONG_PTR", data))
        return self.f2s(data)

    def GetColumnTypes(self, active=True):
        nb = self.info.Control(self._handle(active), self.ffic.FCTL_GETCOLUMNTYPES, 0, 0)
        data = self.ffi.new("wchar_t []", nb)
        self.info.Control(self._handle(active), self.ffic.FCTL_GETCOLUMNTYPES, nb, self.ffi.cast("LONG_PTR", data))
        return self.f2s(data)

    def GetColumnWidths(self, active=True):
        nb = self.info.Control(self._handle(active), self.ffic.FCTL_GETCOLUMNWIDTHS, 0, 0)
        data = self.ffi.new("wchar_t []", nb)
        self.info.Control(self._handle(active), self.ffic.FCTL_GETCOLUMNWIDTHS, nb, self.ffi.cast("LONG_PTR", data))
        return self.f2s(data)

    def GetPanelItem(self, no, active=True):
        # get buffer size
        rc = self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELITEM, no, 0)
        if rc:
            # allocate buffer
            data = self.ffi.new("char []", rc)
            rc = self.info.Control(self._handle(active), self.ffic.FCTL_GETPANELITEM, no, self.ffi.cast("LONG_PTR", data))
            # cast buffer to PluginPanelItem *
            pnli = self.ffi.cast("struct PluginPanelItem *", data)
            return pnli, data
        return None, None

    def GetSelectedPanelItem(self, no, active=True):
        # get buffer size
        rc = self.info.Control(self._handle(active), self.ffic.FCTL_GETSELECTEDPANELITEM, no, 0)
        if rc:
            # allocate buffer
            data = self.ffi.new("char []", rc)
            rc = self.info.Control(self._handle(active), self.ffic.FCTL_GETSELECTEDPANELITEM, no, self.ffi.cast("LONG_PTR", data))
            # cast buffer to PluginPanelItem *
            pnli = self.ffi.cast("struct PluginPanelItem *", data)
            return pnli, data
        return None, None

    def GetCurrentPanelItem(self, active=True):
        # get buffer size
        rc = self.info.Control(self._handle(active), self.ffic.FCTL_GETCURRENTPANELITEM, 0, 0)
        if rc:
            # allocate buffer
            data = self.ffi.new("char []", rc)
            rc = self.info.Control(self._handle(active), self.ffic.FCTL_GETCURRENTPANELITEM, 0, self.ffi.cast("LONG_PTR", data))
            # cast buffer to PluginPanelItem *
            pnli = self.ffi.cast("struct PluginPanelItem *", data)
            return pnli, data
        return None, None

    def BeginSelection(self, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_BEGINSELECTION, 0, 0)

    def SetSelection(self, no, on, active=True):
        on = 1 if on else 0
        return self.info.Control(self._handle(active), self.ffic.FCTL_SETSELECTION, no, self.ffi.cast("LONG_PTR", on))

    def ClearSelection(self, no, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_CLEARSELECTION, no, 0)

    def EndSelection(self, no, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_ENDSELECTION, 0, 0)

    def UpdatePanel(self, keepselection=True, active=True):
        keepselection = 1 if keepselection else 0
        return self.info.Control(self._handle(active), self.ffic.FCTL_UPDATEPANEL, keepselection, 0)

    def RedrawPanel(self, panelinfo=0, active=True):
        return self.info.Control(self._handle(active), self.ffic.FCTL_REDRAWPANEL, 0, self.ffi.cast("LONG_PTR", panelinfo))

    def SetPanelDir(self, paneldir, active=True):
        paneldir = self.s2f(paneldir)
        return self.info.Control(self._handle(active), self.ffic.FCTL_SETPANELDIR, 0, self.ffi.cast("LONG_PTR", paneldir))



#!/usr/bin/env python3

import re
import sys
import xml.etree.ElementTree as ET

import markdown
from markdown.inlinepatterns import LINK_RE, LinkInlineProcessor

[changelog_file, xml_template_file, xml_output_file] = sys.argv[1:]


class StripLinksInlineProcessor(LinkInlineProcessor):
    def handleMatch(self, m, data):
        node, start, end = super().handleMatch(m, data)
        if node is not None:
            node = node.text
        return node, start, end


with open(changelog_file) as fp:
    md_text = fp.read()

md = markdown.Markdown()
md.inlinePatterns.register(
    StripLinksInlineProcessor(LINK_RE, md),
    name="link",
    priority=160,
)

html_content = md.convert(md_text)
html_root = ET.fromstring(f"<root>{html_content}</root>")
release_re = re.compile(r"(?P<version>[\d.]+) (?P<type>[a-z]+) \((?P<date>\d{4}-\d{2}-\d{2})\)")

tree = ET.parse(xml_template_file)
releases = tree.getroot().find("releases")
release_elem = None

for elem in html_root:
    if elem.tag == "h2":
        if elem.text[0].isdecimal():
            match = release_re.match(elem.text)
            release = match.groupdict()
            if release["type"] == "beta":
                release["type"] = "development"
            release_elem = ET.SubElement(releases, "release", release)
    elif elem.tag == "ul" and release_elem is not None:
        for child_elem in elem.iter():
            # https://www.freedesktop.org/software/appstream/docs/chap-Metadata.html#tag-description
            if child_elem.tag == "strong":
                child_elem.tag = "em"
        desc_elem = ET.SubElement(release_elem, "description")
        desc_elem.append(elem)

ET.indent(tree)
tree.write(xml_output_file, encoding="utf-8", xml_declaration=True)



__all__ = ["Window", "Spacer", "HSizer", "VSizer", "GridSizer"]


class Orientation:
    horizontal = 0
    vertical = 1


class Window:
    def __init__(self, pos=(0, 0), size=(0, 0)):
        self.pos = pos
        self.size = size

    def move(self, l, t, w, h):
        self.pos = (l, t)
        self.size = (w, h)

    def get_best_size(self):
        raise NotImplementedError("Window.get_best_size")


class Spacer(Window):
    def get_best_size(self):
        return self.size

class Box:
    def __init__(self, window, border):
        self.window = window
        self.border = border
        assert len(border) == 4

    def size(self, l, t, r, b):
        w = r - l
        h = b - t
        self.window.move(l, t, w, h)

    def get_best_size(self):
        x, y = self.window.get_best_size()
        return (
            x + self.border[0] + self.border[2],
            y + self.border[1] + self.border[3],
        )


class Sizer:
    orientation = None

    def __init__(self, border=(0, 0, 0, 0)):
        assert self.orientation in [Orientation.horizontal, Orientation.vertical]
        self.boxes = []
        self.border = border

    def add(self, w, border=(0, 0, 0, 0)):
        self.boxes.append(Box(w, border))

    def move(self, x, y, w, h):
        self.size(x, y, w, h)

    def size(self, l, t, r, b):
        l += self.border[0]
        t += self.border[1]
        r -= self.border[2]
        b -= self.border[3]
        hoffset = l
        voffset = t
        for box in self.boxes:
            cx, cy = box.get_best_size()
            if self.orientation == Orientation.horizontal:
                box.size(hoffset, voffset, hoffset + cx, voffset + cy)
                hoffset += cx
            elif self.orientation == Orientation.vertical:
                box.size(hoffset, voffset, hoffset + cx, voffset + cy)
                voffset += cy

    def get_best_size(self):
        b_x = 0
        b_y = 0
        for box in self.boxes:
            cx, cy = box.get_best_size()
            if self.orientation == Orientation.horizontal:
                b_x += cx
                b_y = max(cy, b_y)
            elif self.orientation == Orientation.vertical:
                b_x = max(cx, b_x)
                b_y += cy
        return b_x, b_y


class HSizer(Sizer):
    orientation = Orientation.horizontal


class VSizer(Sizer):
    orientation = Orientation.vertical


class FlowSizer(Sizer):
    def __init__(self, cols, border=(0, 0, 0, 0)):
        self.boxes = [[]]
        self.cols = cols
        self.border = border
        self.sizes = None
        self.col_widths = None
        self.row_heights = None

    def add(self, w, border=(0, 0, 1, 0)):
        box = self.boxes[-1]
        if len(box) == self.cols:
            box = []
            self.boxes.append(box)
        box.append(Box(w, border))

    def move(self, x, y, w, h):
        self.size(x, y, w, h)

    def size(self, l, t, r, b):
        l += self.border[0]
        t += self.border[1]
        r -= self.border[2]
        b -= self.border[3]
        cy = 0
        for row in range(len(self.boxes)):
            cx = 0
            for col in range(len(self.boxes[row])):
                box = self.boxes[row][col]
                bx, by = box.get_best_size()
                box.size(l+cx, t+cy, l+cx+self.col_widths[col], t+cy+by)
                cx += self.col_widths[col]
            cy += self.row_heights[row]

    def get_best_size(self):
        while len(self.boxes[-1])%self.cols:
            self.add(Spacer())
        sizes = [[col.get_best_size() for col in row] for row in self.boxes]
        col_widths = [max(sizes[row_nr][col_nr][0] for row_nr in range(len(sizes))) for col_nr in range(len(sizes[0]))]
        row_heights = [max(col[1] for col in row) for row in sizes]
        self.sizes = sizes
        self.col_widths = col_widths
        self.row_heights = row_heights
        w = sum(col_widths)
        h = sum(row_heights)
        return w, h



import logging
import time
import ctypes as ct
from far2l.plugin import PluginBase
import far2lc


log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python ED Indent"
    openFrom = ["PLUGINSMENU", "EDITOR"]

    def OpenPlugin(self, OpenFrom):
        return 0

    def ProcessEditorInput(self, Rec):
        rec = self.ffi.cast("INPUT_RECORD *", Rec)

        if (
            rec.EventType != self.ffic.KEY_EVENT
            or not rec.Event.KeyEvent.bKeyDown
            or (rec.Event.KeyEvent.wVirtualKeyCode != self.ffic.VK_TAB)
            or (rec.Event.KeyEvent.dwControlKeyState & (
                self.ffic.RIGHT_ALT_PRESSED
                | self.ffic.LEFT_ALT_PRESSED
                | self.ffic.RIGHT_CTRL_PRESSED
                | self.ffic.LEFT_CTRL_PRESSED
            ))
        ):
            return 0
        indent = (rec.Event.KeyEvent.dwControlKeyState & (self.ffic.SHIFT_PRESSED)) != 0
        self.Perform(indent)
        return 1

    def Perform(self, indent):
        ei = self.ffi.new("struct EditorInfo *")
        self.info.EditorControl(self.ffic.ECTL_GETINFO, ei)

        if ei.BlockType != self.ffic.BTYPE_STREAM:
            # working only if not vertical block selection
            return 0;

        egs = self.ffi.new("struct EditorGetString *")
        egs.StringNumber = ei.BlockStartLine
        self.info.EditorControl(self.ffic.ECTL_GETSTRING, egs)
        if egs.SelEnd != -1:
            return 0

        eur = self.ffi.new("struct EditorUndoRedo *")
        eur.Command = self.ffic.EUR_BEGIN
        self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)

        IndentStr = '\t'
        fIndentStr = self.s2f(IndentStr)

        for lno in range(ei.BlockStartLine, ei.TotalLines):
            esp = self.ffi.new("struct EditorSetPosition *")
            esp.CurLine = lno
            esp.CurPos = esp.Overtype = 0
            esp.CurTabPos = esp.TopScreenLine = esp.LeftPos = -1
            self.info.EditorControl(self.ffic.ECTL_SETPOSITION, esp)

            egs = self.ffi.new("struct EditorGetString *")
            egs.StringNumber = -1
            self.info.EditorControl(self.ffic.ECTL_GETSTRING, egs)

            if egs.SelStart == -1 or egs.SelStart == egs.SelEnd:
                # Stop when reaching the end of the text selection
                break

            if egs.SelEnd != -1:
                # if selection in line not up to end we force selection block up to line end
                # because if selection not up to end the tab replace selected text in last line
                es = self.ffi.new("struct EditorSelect *")
                es.BlockType = ei.BlockType
                es.BlockStartLine = ei.BlockStartLine
                es.BlockStartPos = 0
                es.BlockWidth = 0
                es.BlockHeight = line - ei.BlockStartLine + 1
                self.info.EditorControl(self.ffic.ECTL_SELECT, es)

            if indent:
                self.info.EditorControl(self.ffic.ECTL_INSERTTEXT, fIndentStr)
            else:
                if egs.StringText[0] == '\t':
                    n = 1
                else:
                    n = 0
                    s = self.f2s(egs.StringText)
                    while n < ei.TabSize and s[n] == ' ':
                        n += 1
                for i in range(n):
                    self.info.EditorControl(self.ffic.ECTL_DELETECHAR, self.ffi.NULL)
            
        eur = self.ffi.new("struct EditorUndoRedo *")
        eur.Command = self.ffic.EUR_END
        self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)

        self.info.EditorControl(self.ffic.ECTL_REDRAW, self.ffi.NULL)



import logging
import uuid
from far2l.plugin import PluginBase


log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python uuuidgen"
    openFrom = ["PLUGINSMENU", "EDITOR"]

    @staticmethod
    def HandleCommandLine(line):
        return line in ("uuidgen",)

    def CommandLine(self, line):
        if line != "uuidgen":
            return
        self.uuidgen()

    def uuidgen(self):
        s = '{%s}'%uuid.uuid4()
        log.debug('uuidgen: {}'.format(s))
        self.info.FSF.CopyToClipboard(self.s2f(s))

    def OpenPlugin(self, OpenFrom):
        self.uuidgen()
        return -1



import os
import logging

import far2lc
from far2l.plugin import PluginBase

log = logging.getLogger(__name__)

class Plugin(PluginBase):
    label = "Python Clip GET"
    openFrom = ["PLUGINSMENU", 'FILEPANEL']

    def CopyFiles(self, data):
        pwd = self.panel.GetPanelDir()
        data = data.decode('utf8')
        log.debug(f'copyfiles: {data}')
        prefix = 'file://'
        for uri in data.split('\n'):
            if uri[:len(prefix)] != prefix:
                continue
            fqname = uri[len(prefix):]
            fname = fqname.split('/')[-1]
            dqname = os.path.join(pwd, fname)
            log.debug(f'CopyFile: {fqname} -> {dqname}')
            with open(dqname, 'wb') as fo:
                with open(fqname, 'rb') as fi:
                    while True:
                        rec = fi.read(4096)
                        if not rec:
                            break
                        fo.write(rec)

    def OpenPlugin(self, OpenFrom):
        winport = self.ffi.cast("struct WINPORTDECL *", far2lc.WINPORT())
        clipurifmt = winport.RegisterClipboardFormat("text/uri-list")
        if not clipurifmt:
            log.error('uclipset.ClipboardRegisterFormat.1')
            return
        clipgnofmt = winport.RegisterClipboardFormat("x-special/gnome-copied-files")
        if not clipgnofmt:
            log.error('uclipset.ClipboardRegisterFormat.2')
            return
        if not winport.OpenClipboard(self.ffi.NULL):
            log.error('uclipset.OpenClipboard')
            return
        try:
            data = winport.GetClipboardData(clipurifmt)
            if data is not None:
                nb = winport.ClipboardSize(data)
                result = self.ffi.buffer(self.ffi.cast("PSTR", data), nb)
                self.CopyFiles(bytes(result))
            else:
                data = winport.GetClipboardData(clipgnofmt)
                if data is not None:
                    nb = winport.ClipboardSize(data)
                    result = self.ffi.buffer(self.ffi.cast("PSTR", data), nb)
                    self.CopyFiles(bytes(result))
        except:
            log.exception('uclipset.GetClipboardData')
        finally:
            if not winport.CloseClipboard():
                log.error('uclipset.CloseClipboard')
                return



import shlex
import logging
import optparse
from far2l.plugin import PluginBase
from far2l import fardialogbuilder as dlgb
from far2lpackages import googletranslate

log = logging.getLogger(__name__)


class OptionParser(optparse.OptionParser):
    def error(self, msg):
        log.error("utranslate: {}".format(msg))
        raise ValueError("bad option")


class Plugin(PluginBase):
    label = "Python utranslate"
    openFrom = ["PLUGINSMENU", "EDITOR"]

    @staticmethod
    def HandleCommandLine(line):
        return line in ("translate", "spell")

    def CommandLine(self, line):
        cmd, line = line.split(" ", 1)
        if cmd == "spell":
            args = shlex.split(line)
            parser = OptionParser()
            parser.add_option("--lang", action="store", dest="lang", default="en")
            parser.add_option("--text", action="store", dest="text", default="en")
            try:
                opts, args = parser.parse_args(args)
            except ValueError:
                return
            log.debug("spell: opts={} args={}".format(opts, args))
            try:
                t = googletranslate.get_translate(opts.text, opts.lang, opts.lang)
            except:
                log.exception("spell")
                return
            log.debug("spell result: {}".format(t))
        elif cmd == "translate":
            args = shlex.split(line)
            parser = optparse.OptionParser()
            parser.add_option("--from", action="store", dest="lfrom", default="en")
            parser.add_option("--to", action="store", dest="lto", default="en")
            parser.add_option("--text", action="store", dest="text", default="")
            try:
                opts, args = parser.parse_args(args)
            except ValueError:
                return
            log.debug("translate: opts={} args={}".format(opts, args))
            try:
                t = googletranslate.get_translate(opts.text, opts.lfrom, opts.lto)
            except:
                log.exception("translate")
                return
            log.debug("translate result: {}".format(t))
        else:
            log.error("bad command:{} line={}".format(cmd, line))
            return
        s = t["result"][0]
        self.info.FSF.CopyToClipboard(self.s2f(s))

    def Dialog(self):
        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                try:
                    s = self.info.FSF.PasteFromClipboard()
                    s = self.f2s(s)
                except:
                    s = "And do beautiful things"
                try:
                    dlg.SetText(dlg.ID_text, s)
                    dlg.SetCheck(dlg.ID_spell, 1)
                    dlg.SetText(dlg.ID_from, "en")
                    dlg.SetText(dlg.ID_to, "pl")
                    dlg.SetFocus(dlg.ID_text)
                except:
                    log.exception("Init Dialog")
            elif Msg == self.ffic.DN_BTNCLICK:
                if dlg.ID_perform == Param1:
                    text = dlg.GetText(dlg.ID_text)
                    spell = dlg.GetCheck(dlg.ID_spell)
                    translate = dlg.GetCheck(dlg.ID_translate)
                    lfrom = dlg.GetText(dlg.ID_from)
                    lto = dlg.GetText(dlg.ID_to)
                    log.debug(
                        "DN_BTNCLICK(spell={}, translate={} nfrom={} nto={} text={})".format(
                            spell, translate, lfrom, lto, text
                        )
                    )
                    try:
                        if spell:
                            t = googletranslate.get_translate(text, lfrom, lfrom)
                        else:
                            t = googletranslate.get_translate(text, lfrom, lto)
                        log.debug("translate result: {}".format(t))
                        s = t["result"][0]
                        self.info.FSF.CopyToClipboard(self.s2f(s))
                    except:
                        log.exception("translate")
                        s = "connection error"
                    dlg.SetText(dlg.ID_result, s)
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        b = dlgb.DialogBuilder(
            self,
            DialogProc,
            "Python utranslate",
            "helptopic",
            0,
            dlgb.VSizer(
                dlgb.HSizer(
                    dlgb.TEXT(None, "Text:"),
                    dlgb.Spacer(),
                    dlgb.EDIT("text", 60, maxlength=120),
                ),
                dlgb.HSizer(
                    dlgb.TEXT(None, "Operation:"),
                    dlgb.Spacer(),
                    dlgb.RADIOBUTTON("spell", "Spell", True, flags=self.ffic.DIF_GROUP),
                    dlgb.Spacer(),
                    dlgb.RADIOBUTTON("translate", "Translate"),
                ),
                dlgb.HSizer(
                    dlgb.TEXT(None, "From:"),
                    dlgb.Spacer(),
                    dlgb.EDIT("from", 4, maxlength=4),
                ),
                dlgb.HSizer(
                    dlgb.TEXT(None, "To:"),
                    dlgb.Spacer(),
                    dlgb.EDIT("to", 4, maxlength=4),
                ),
                dlgb.HLine(),
                dlgb.HSizer(
                    dlgb.TEXT(None, "Result:"),
                    dlgb.Spacer(),
                    dlgb.EDIT("result", 60, maxlength=120),
                ),
                dlgb.HLine(),
                dlgb.HSizer(
                    dlgb.BUTTON(
                        "perform",
                        "Perform",
                        default=True,
                        flags=self.ffic.DIF_CENTERGROUP | self.ffic.DIF_BTNNOCLOSE,
                    ),
                    dlgb.BUTTON("close", "Close", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        log.debug("translate res={0}".format(res))

        self.info.DialogFree(dlg.hDlg)

    def OpenPlugin(self, OpenFrom):
        if OpenFrom == 5:
            # EDITOR
            self.Dialog()
            return -1
        _MsgItems = [
            self.s2f("Python Translate/Spellcheck from="),
            self.s2f(""),
            self.s2f("From shell:"),
            self.s2f(
                "  py:translate --from=en --to=pl --text='And do beautiful things'"
            ),
            self.s2f("  py:spell --lang=en --text='And do beautifull things'"),
            self.s2f("From editor:"),
            self.s2f("  copy text to clipboard, execute plugin"),
            self.s2f(""),
            self.s2f("Result is in clipboard :)"),
            self.s2f(""),
            self.s2f("\x01"),
            self.s2f("&Ok"),
        ]
        MsgItems = self.ffi.new("wchar_t *[]", _MsgItems)
        self.info.Message(
            self.info.ModuleNumber,  # GUID
            self.ffic.FMSG_WARNING | self.ffic.FMSG_LEFTALIGN,  # Flags
            self.s2f("Contents"),  # HelpTopic
            MsgItems,  # Items
            len(MsgItems),  # ItemsNumber
            1,  # ButtonsNumber
        )
        return -1



import logging


log = logging.getLogger(__name__)


class Dialog:
    def __init__(self, plugin):
        self.info = plugin.info
        self.ffic = plugin.ffic
        self.ffi = plugin.ffi
        self.s2f = plugin.s2f
        self.f2s = plugin.f2s

        self.width = 0
        self.height = 0
        self.dialogItems = []
        # self.id2item = {}
        self.fdi = None
        self.hDlg = None

    def Close(self, exitcode):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_CLOSE, exitcode, 0)

    def EnableRedraw(self, on):
        self.info.SendDlgMessage(
            self.hDlg, self.ffic.DM_ENABLEREDRAW, 1 if on else 0, 0
        )

    def RedrawDialog(self):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_REDRAW, 0, 0)

    def GetDlgData(self):
        return self.info.SendDlgMessage(self.hDlg, self.ffic.DM_GETDLGDATA, 0, 0)

    def SetDlgData(self, Data):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_SETDLGDATA, 0, Data)

    def GetDlgItemData(self, ID):
        return self.info.SendDlgMessage(self.hDlg, self.ffic.DM_GETITEMDATA, ID, 0)

    def SetDlgItemData(self, ID, Data):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_SETITEMDATA, ID, Data)

    def GetFocus(self, hDlg):
        return self.info.SendDlgMessage(self.hDlg, self.ffic.DM_GETFOCUS, 0, 0)

    def SetFocus(self, ID):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_SETFOCUS, ID, 0)

    def Enable(self, ID):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_ENABLE, ID, 1)

    def Disable(self, ID):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_ENABLE, ID, 0)

    def IsEnable(self, ID):
        return self.info.SendDlgMessage(self.hDlg, self.ffic.DM_ENABLE, ID, -1)

    def GetTextLength(self, ID):
        return self.info.SendDlgMessage(self.hDlg, self.ffic.DM_GETTEXTPTR, ID, 0)

    def GetText(self, ID):
        sptr = self.info.SendDlgMessage(self.hDlg, self.ffic.DM_GETCONSTTEXTPTR, ID, 0)
        return self.f2s(sptr)

    def SetText(self, ID, Str):
        sptr = self.s2f(Str)
        # preserve item.data ?
        # self.id2item[ID][1] = sptr
        self.info.SendDlgMessage(
            self.hDlg, self.ffic.DM_SETTEXTPTR, ID, self.ffi.cast("LONG_PTR", sptr)
        )

    def GetCheck(self, ID):
        return self.info.SendDlgMessage(self.hDlg, self.ffic.DM_GETCHECK, ID, 0)

    def SetCheck(self, ID, State):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_SETCHECK, ID, State)

    def AddHistory(self, ID, Str):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_ADDHISTORY, ID, Str)

    def AddString(self, ID, Str):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_LISTADDSTR, ID, Str)

    def GetCurPos(self, ID):
        return self.info.SendDlgMessage(self.hDlg, self.ffic.DM_LISTGETCURPOS, ID, 0)

    # def SetCurPos(self, ID, NewPos):
    #    struct FarListPos LPos={NewPos, -1}
    #    self.info.SendDlgMessage(self.hDlg, self.ffic.DM_LISTSETCURPOS, ID, (LONG_PTR)&LPos)

    def ClearList(self, ID):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_LISTDELETE, ID, 0)

    # def DeleteItem(self, ID, Index):
    #    struct FarListDelete FLDItem={Index, 1}
    #    self.info.SendDlgMessage(self.hDlg, self.ffic.DM_LISTDELETE, ID, (LONG_PTR)&FLDItem)

    def SortUp(self, ID):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_LISTSORT, ID, 0)

    def SortDown(self, ID):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_LISTSORT, ID, 1)

    def GetItemData(self, ID, Index):
        return self.info.SendDlgMessage(self.hDlg, self.ffic.DM_LISTGETDATA, ID, Index)

    # def SetItemStrAsData(self, ID, Index, Str):
    #    struct FarListItemData FLID{Index, 0, Str, 0}
    #    self.info.SendDlgMessage(self.hDlg, self.ffic.DM_LISTSETDATA, ID, (LONG_PTR)&FLID)

    def GetCursorPos(self, ID):
        cpos = self.ffi.new("COORD *")
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_GETCURSORPOS, ID, self.ffi.cast("LONG_PTR", cpos))
        return (cpos.X, cpos.Y)

    def SetCursorPos(self, ID, X, Y):
        cpos = self.ffi.new("COORD *", dict(X=X, Y=Y))
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_SETCURSORPOS, ID, self.ffi.cast("LONG_PTR", cpos))

    def ShowItem(self, ID, On):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_SHOWITEM, ID, On)

    def SetCursorSize(self, ID, On, Size):
        self.info.SendDlgMessage(self.hDlg, self.ffic.DM_SETCURSORSIZE, ID, On|(Size<<16))



import logging
from far2l.plugin import PluginBase
from far2l.fardialogbuilder import (
    TEXT,
    EDIT,
    BUTTON,
    USERCONTROL,
    HLine,
    HSizer,
    VSizer,
    DialogBuilder,
)


log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python Character Map"
    openFrom = ["PLUGINSMENU", "COMMANDLINE", "EDITOR", "VIEWER", "DIALOG"]

    def OpenPlugin(self, OpenFrom):
        try:
            return self._OpenPlugin(OpenFrom)
        except Exception as ex:
            log.exception('run')

    def GetCursorPos(self, hDlg, ID):
        cpos = self.ffi.new("COORD *")
        self.info.SendDlgMessage(hDlg, self.ffic.DM_GETCURSORPOS, ID, self.ffi.cast("LONG_PTR", cpos))
        return (cpos.X, cpos.Y)

    def SetCursorPos(self, hDlg, ID, col, row):
        cpos = self.ffi.new("COORD *", dict(X=col, Y=row))
        self.info.SendDlgMessage(hDlg, self.ffic.DM_SETCURSORPOS, ID, self.ffi.cast("LONG_PTR", cpos))

    def _OpenPlugin(self, OpenFrom):
        def GetColor(no):
            data = self.ffi.new("DWORD *")
            rc = self.info.AdvControl(
                self.info.ModuleNumber,
                self.ffic.ACTL_GETCOLOR,
                self.ffi.cast("VOID *", no),
                self.ffi.cast("VOID *", data)
            )
            return data[0]

        self.offset = 0
        self.max_col = 64
        self.max_row = 20
        self.text = ''
        self.charbuf = self.ffi.new("CHAR_INFO []", self.max_col*self.max_row)
        attrNormal = GetColor(self.ffic.COL_EDITORTEXT)
        attrSelected = GetColor(self.ffic.COL_EDITORSELECTEDTEXT)
        #attrNormal = 0x170
        #attrSelected = 0x1c

        def setVBuf(hDlg):
            self.info.SendDlgMessage(hDlg, self.ffic.DM_ENABLEREDRAW, 0, 0)
            for i in range(self.max_col*self.max_row):
                cb = self.charbuf[i]
                cb.Attributes = attrNormal
                if self.offset+i <= 0xffff:
                    cb.Char.UnicodeChar = self.offset+i
                else:
                    cb.Char.UnicodeChar = 0x20
            self.info.SendDlgMessage(hDlg, self.ffic.DM_ENABLEREDRAW, 1, 0)

        def xy2lin(col, row):
            return row * self.max_col + col

        def setColor(col, row, attr):
            self.charbuf[xy2lin(col, row)].Attributes = attr

        def updateOffset(offset):
            dlg.SetText(dlg.ID_voffset, f"{offset:5d}/0x{offset:04x}")

        def _DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                setVBuf(hDlg)
                self.SetCursorPos(hDlg, dlg.ID_hex, 0, 0)
                setColor(0, 0, attrSelected)
                updateOffset(0)
                self.info.SendDlgMessage(hDlg, self.ffic.DM_SETCURSORSIZE, dlg.ID_hex, 0|(0<<16))
                dlg.SetFocus(dlg.ID_hex)
                return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)
            elif Msg == self.ffic.DN_BTNCLICK:
                #log.debug(f"btn DialogProc({Param1}, {Param2})")
                if Param1 == dlg.ID_vgoto:
                    v = dlg.GetText(dlg.ID_vgotovalue).strip()
                    if v:
                        try:
                            offset = int(v, 16 if v[:2] == '0x' else 10)
                            if offset > 0xffff:
                                offset = 0
                                dlg.SetText(dlg.ID_vgotovalue, '0')
                            self.offset = offset
                        except:
                            if len(v) == 1:
                                self.offset = ord(v)
                                setVBuf(hDlg)
                            else:
                                dlg.SetText(dlg.ID_vgotovalue, '')
                                log.exception('goto')
                    setVBuf(hDlg)
                    updateOffset(self.offset)
                    col, row = self.GetCursorPos(hDlg, dlg.ID_hex)
                    setColor(col, row, attrNormal)
                    setColor(0, 0, attrSelected)
                    self.SetCursorPos(hDlg, dlg.ID_hex, 0, 0)
                    dlg.SetFocus(dlg.ID_hex)
                    return 1
                elif Param1 == dlg.ID_vok:
                    col, row = self.GetCursorPos(hDlg, dlg.ID_hex)
                    offset = self.offset + xy2lin(col, row)
                    if offset <= 0xffff:
                        try:
                            self.text = chr(offset)
                        except:
                            log.exception(f'OK click {offset}')
                return 0
            elif Msg == self.ffic.DN_KEY and Param1 == dlg.ID_hex:
                col, row = self.GetCursorPos(hDlg, dlg.ID_hex)
                setColor(col, row, attrNormal)
                #log.debug(f"key DialogProc({Param1}, {Param2:x}), col={col} row={row})")
                if Param2 == self.ffic.KEY_PGUP:
                    self.offset -= self.max_row * self.max_col
                    if self.offset < 0:
                        self.offset = 0
                    setVBuf(hDlg)
                    offset = self.offset + xy2lin(col, row)
                    updateOffset(offset)
                elif Param2 == self.ffic.KEY_HOME:
                    row = 0
                    col = 0
                elif Param2 == self.ffic.KEY_PGDN:
                    self.offset += self.max_row * self.max_col
                    if self.offset > 0xffff:
                        self.offset = 0x10000 - self.max_row * self.max_col
                    setVBuf(hDlg)
                    offset = self.offset + xy2lin(col, row)
                    updateOffset(offset)
                elif Param2 == self.ffic.KEY_LEFT:
                    col -= 1
                elif Param2 == self.ffic.KEY_UP:
                    row -= 1
                elif Param2 == self.ffic.KEY_RIGHT:
                    col += 1
                elif Param2 == self.ffic.KEY_DOWN:
                    row += 1
                elif Param2 == self.ffic.KEY_ENTER:
                    #log.debug(f"enter at offset:{self.offset} row:{row} col:{col}")
                    offset = self.offset + xy2lin(col, row)
                    if offset <= 0xffff:
                        try:
                            self.text = chr(offset)
                        except:
                            log.exception(f'enter at {offset}')
                    return 0
                elif Param2 == self.ffic.KEY_ESC:
                    return 0
                else:
                    return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)
                if col == self.max_col:
                    col = 0
                elif col == -1:
                    col = self.max_col - 1
                if row == self.max_row:
                    row = 0
                elif row == -1:
                    row = self.max_row - 1
                setColor(col, row, attrSelected)
                offset = self.offset + xy2lin(col, row)
                updateOffset(offset)
                #log.debug(f"col={col}/{self.max_col} row={row}/{self.max_row}")
                self.SetCursorPos(hDlg, dlg.ID_hex, col, row)
                return 1
            elif Msg == self.ffic.DN_MOUSECLICK and Param1 == dlg.ID_hex:
                col, row = self.GetCursorPos(hDlg, dlg.ID_hex)
                setColor(col, row, attrNormal)
                mou = self.ffi.cast("MOUSE_EVENT_RECORD *", Param2)
                col = mou.dwMousePosition.X
                row = mou.dwMousePosition.Y
                setColor(col, row, attrSelected)
                self.SetCursorPos(hDlg, dlg.ID_hex, col, row)
                #log.debug(f"mou DialogProc(col={col} row={row})")
                offset = self.offset + xy2lin(col, row)
                #self.text = self.symbols[offset]
                updateOffset(offset)
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            try:
                return _DialogProc(hDlg, Msg, Param1, Param2)
            except:
                log.exception('dialogproc')
                return 1

        b = DialogBuilder(
            self,
            DialogProc,
            "Python CharMap",
            "charmap",
            0,
            VSizer(
                HSizer(
                    TEXT(None, "Offset:"),
                    TEXT("voffset", " "*12),
                    TEXT(None, "&Goto:"),
                    EDIT("vgotovalue", width=7),
                    BUTTON("vgoto", "GOTO"),
                ),
                HLine(),
                USERCONTROL('hex', self.max_col, self.max_row, param={'VBuf':self.ffi.cast("CHAR_INFO *", self.ffi.addressof(self.charbuf))}),
                HLine(),
                HSizer(
                    BUTTON("vok", "OK", flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("vcancel", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        #log.debug(f'res={res} text=***{self.text}***')
        if res in (dlg.ID_hex, dlg.ID_vok):
            self.info.FSF.CopyToClipboard(self.s2f(self.text))
        self.info.DialogFree(dlg.hDlg)



import os
import re

import cffi
ffi = cffi.FFI()

__all__ = ["ffi", "ffic"]

def convert():
    with open(os.path.join(os.path.dirname(__file__), 'far2lcffi.h'), 'rt') as fp:
        data = fp.read()

    for s in re.findall("'.+'", data):
        i = str(ord(s[-2].encode("ascii")))
        data = data.replace(s, i)

    pack = None
    while True:
        i = data.find('#pragma pack')
        if i < 0:
            ffi.cdef(data, pack=pack)
            break
        ndata = data[:i]
        ffi.cdef(ndata, pack=pack)
        data = data[i:]
        i = data.find(')')
        s = data[:i+1]
        if s.find('()') > 0:
            pack = None
        else:
            pack = 2
        data = data[i+1:]

convert()

ffi.cdef('''
wchar_t *wcsdup(wchar_t *src);
void *calloc(size_t size, size_t count);
void *malloc(size_t size);
void free(void *ptr);
''')

ffic = ffi.dlopen(None)



#!python3
import sys

def filter_by_word (line):
	bad_words = [
		'cosmetic', 'minor fix', 'refactor', 'Merge pull request', 
		'cleanup', 'comment', 'Merge branch', 'formatting', 'minor', 
		'commit missing', 'debugging', 'logging', 'naming']
	for x in bad_words:
		if x in line:
			return True
	return False

dates_list = []
dates_map = {}
date = ''

while True:
	try:
		line = input()
		line = line.strip()
		if line.startswith('*'):
			items = line.split();
			date = items[1] + ' ' + items[2] + ' ' + items[3] + ' ' + items[4]
			if '2022' in items[4]:
				dates_list.append(date)
				dates_map[date] = [line, '- much more development from 2015 - 2022 years']
				break
			if not date in dates_map:
				dates_list.append(date)
				dates_map[date] = [line]
		elif line.startswith('-'):
			if not filter_by_word(line):
				dates_map[date].append(line)

	except EOFError:
		break

for date in dates_list:
	lines = dates_map[date]
	if len(lines) > 1:
		for y in lines:
			print(y)
		print ()


import logging
import time
import ctypes as ct
from far2l.plugin import PluginBase
import far2lc


log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python ED Uniq"
    openFrom = ["PLUGINSMENU", "EDITOR"]

    def Perform(self):
        ei = self.ffi.new("struct EditorInfo *")
        self.info.EditorControl(self.ffic.ECTL_GETINFO, ei)

        if ei.BlockType != self.ffic.BTYPE_STREAM:
            # working only if not vertical block selection
            return 0;

        egs = self.ffi.new("struct EditorGetString *")
        egs.StringNumber = ei.BlockStartLine
        self.info.EditorControl(self.ffic.ECTL_GETSTRING, egs)
        if egs.SelEnd != -1:
            return 0

        eur = self.ffi.new("struct EditorUndoRedo *")
        eur.Command = self.ffic.EUR_BEGIN
        self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)

        lines = []
        result = []
        for lno in range(ei.BlockStartLine, ei.TotalLines):
            esp = self.ffi.new("struct EditorSetPosition *")
            esp.CurLine = lno
            esp.CurPos = esp.Overtype = 0
            esp.CurTabPos = esp.TopScreenLine = esp.LeftPos = -1
            self.info.EditorControl(self.ffic.ECTL_SETPOSITION, esp)

            egs = self.ffi.new("struct EditorGetString *")
            egs.StringNumber = -1
            self.info.EditorControl(self.ffic.ECTL_GETSTRING, egs)

            if egs.SelStart == -1 or egs.SelStart == egs.SelEnd:
                # Stop when reaching the end of the text selection
                break

            if egs.SelEnd != -1:
                # Extend selection to line end
                es = self.ffi.new("struct EditorSelect *")
                es.BlockType = ei.BlockType
                es.BlockStartLine = ei.BlockStartLine
                es.BlockStartPos = 0
                es.BlockWidth = 0
                es.BlockHeight = line - ei.BlockStartLine + 1
                self.info.EditorControl(self.ffic.ECTL_SELECT, es)

            s = self.f2s(egs.StringText)
            if s not in lines:
                lines.append(s)

        self.info.EditorControl(self.ffic.ECTL_DELETEBLOCK, self.ffi.NULL)

        s = '\r'.join(lines)+'\r'
        s = self.s2f(s)
        self.info.EditorControl(self.ffic.ECTL_INSERTTEXT, s)

        eur = self.ffi.new("struct EditorUndoRedo *")
        eur.Command = self.ffic.EUR_END
        self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)


    def OpenPlugin(self, OpenFrom):
        if OpenFrom == 5:
            # EDITOR
            self.Perform()
        return -1



import time
import logging
from far2l.plugin import PluginBase
from far2l.farprogress import ProgressMessage

log = logging.getLogger(__name__)

class Plugin(PluginBase):
    label = "Python Progress Message"
    openFrom = ["PLUGINSMENU"]

    def OpenPlugin(self, OpenFrom):
        t = ProgressMessage(self, "Progress demo", "Please wait ... working", 100)
        t.show()
        time.sleep(2)
        for i in range(0, 100, 20):
            if t.aborted():
                break
            t.update(i)
            time.sleep(2)
        t.close()




"""
This is a work-in-progress library which tries
to make writing FAR plugins as simple as possible by hiding
all of the C API behind pythonic abstractions.

Everything that is not underscored is for public usage,
but until majority of FAR API will be accessible via
this library, interfaces will be changed frequently.
"""

__author__ = 'Yaroslav Yanovsky'

import logging
import os

from far2l.plugin import PluginBase

_log = logging.getLogger(__name__)


class _File:
    """
    Class that represents file on FAR panel.
    Do not create directly, only use properites.
    """
    def __init__(self, panel, data, index=None, is_selected=None):
        """
        ``data`` is a FAR_FIND_DATA to build file info around.
        ``index`` is a position of the file on panel.
        """
        self._panel = panel
        self.creation_time = data.ftCreationTime
        self.last_access_time = data.ftLastAccessTime
        self.last_write_time = data.ftLastWriteTime
        self.physical_size = data.nPhysicalSize
        self.file_size = data.nFileSize
        self.file_attributes = data.dwFileAttributes
        self.unix_mode = data.dwUnixMode
        self.file_name = panel._plugin.f2s(data.lpwszFileName)
        self._index = index
        self._is_selected = is_selected

    @property
    def full_file_name(self):
        """
        Tries to provide full file name if possible.
        """
        return os.path.join(self._panel.directory, self.file_name)

    @property
    def index(self):
        """
        Returns file position on the panel.
        """
        if self._index is not None:
            return self._index
        # index not provided on creation - scan through panel to find it.
        for f in self._panel:
            if f.file_name == self.file_name:
                self._index = f.index
                return f.index
        raise ValueError('Unable to determine index for file {!r}'.format(
            self.file_name))

    @property
    def is_selected(self):
        """
        Returns whether file is selected or not.
        """
        if self._is_selected is None:
            # selection status  not provided on creation - fetch this 
            # info from file panel
            self._is_selected = (self.file_name 
                                 in self._panel._get_selection_dict())
        return self._is_selected


class _Panel:
    """
    Class that represents FAR panel.
    Do not create directly, only use properites.
    """
    def __init__(self, plugin, is_active):
        self._plugin = plugin
        self._handle = plugin.ffi.cast('HANDLE', -1 if is_active else -2)

        PPI = (self._get_control_sized, 'struct PluginPanelItem *')
        STR = (self._get_control_str, None)
        PI = (self._get_control_fixed, 'struct PanelInfo *')
        c = plugin.ffic
        self._CONTROL_RETURN_TYPES = {
            c.FCTL_GETPANELITEM: PPI,
            c.FCTL_GETCURRENTPANELITEM: PPI,
            c.FCTL_GETSELECTEDPANELITEM: PPI,
            c.FCTL_GETPANELINFO: PI,
            c.FCTL_GETPANELDIR: STR,
            c.FCTL_GETPANELFORMAT: STR,
            c.FCTL_GETPANELHOSTFILE: STR,
        }

    @property
    def directory(self):
        """
        Returns current panel directory.
        """
        return self._get_control(self._plugin.ffic.FCTL_GETPANELDIR)

    @property
    def cursor(self):
        """
        Returns index of file under cursor.
        """
        return self._get_info().CurrentItem

    @cursor.setter
    def cursor(self, index):
        """
        Changes file cursor - ``index`` is a file index in the panel.
        """
        if index < 0 or index >= len(self):
            raise IndexError('Out of range')
        redraw = self._plugin.ffi.new('struct PanelRedrawInfo *')
        redraw.CurrentItem = index
        redraw.TopPanelItem = -1
        self._plugin.info.Control(self._handle, 
                                  self._plugin.ffic.FCTL_REDRAWPANEL, 0,
                                  self._plugin.ffi.cast('LONG_PTR', redraw))

    @property
    def current(self):
        """
        Returns current file.
        """
        return self[self.cursor]
    
    def __len__(self):
        """
        Returns total amount of files on the panel.
        """
        return self._get_info().ItemsNumber

    def __getitem__(self, index):
        """
        Returns panel file at given ``index``.
        """
        if index < 0 or index >= len(self):
            raise IndexError('Out of range')
        return self._get_item_impl(index)

    def _get_item_impl(self, index, selection_dict=None):
        item = self._get_control(self._plugin.ffic.FCTL_GETPANELITEM, index)
        f = _File(self, item.FindData, index)
        if selection_dict is not None:
            f._is_selected = f.file_name in selection_dict
        return f

    @property
    def selected(self):
        """
        Returns list of selected files on the panel.
        """
        res = []
        for i in range(self._get_info().SelectedItemsNumber):
            item = self._get_control(
                self._plugin.ffic.FCTL_GETSELECTEDPANELITEM, i)
            res.append(_File(self, item.FindData, is_selected=True))
        return res

    def _get_selection_dict(self):
        return {item.file_name: item for item in self.selected}

    def __iter__(self):
        """
        Iterate over all files present on current panel.
        """
        selection = self._get_selection_dict()
        for i in range(len(self)):
            yield self._get_item_impl(i, selection)

    def refresh(self):
        """
        Updates and redraws this panel.
        """
        p = self._plugin
        p.info.Control(self._handle, p.ffic.FCTL_UPDATEPANEL, 0, 0)
        p.info.Control(self._handle, p.ffic.FCTL_REDRAWPANEL, 0, 0)

    def _get_control_str(self, _, command, arg=0):
        p = self._plugin
        char_count = p.info.Control(self._handle, command, 0, 0)
        data = p.ffi.new('wchar_t []', char_count)
        p.info.Control(self._handle, command, char_count, 
                       p.ffi.cast('LONG_PTR', data))
        return p.f2s(data)
        
    def _get_control_sized(self, cast_type, command, arg=0):
        p = self._plugin
        size = p.info.Control(self._handle, command, arg, 0)
        if size:
            data = p.ffi.new('char []', size)
            p.info.Control(self._handle, command, arg, 
                           p.ffi.cast('LONG_PTR', data))
            return p.ffi.cast(cast_type, data)
        return None

    def _get_control_fixed(self, cast_type, command, arg=0):
        p = self._plugin
        data = p.ffi.new(cast_type)
        if p.info.Control(self._handle, command, arg, 
                          p.ffi.cast('LONG_PTR', data)):
            return data
        return None

    def _get_control(self, command, arg=0):
        func, cast_type = self._CONTROL_RETURN_TYPES[command]
        result = func(cast_type, command, arg)
        if result is None:
            _log.error('Control({}, {}, {}, ...) failed, expecting {}'.format(
                self._handle, command, arg, cast_type))
        return result

    def _get_info(self):
        return self._get_control(self._plugin.ffic.FCTL_GETPANELINFO)


class _Editor:
    def __init__(self, plugin):
        self._plugin = plugin
        self._control = self._plugin.info.EditorControl
        self._info = None

    def _get_info(self):
        info = self._info
        if info is None:  # cache the info
            p = self._plugin
            info = p.ffi.new('struct EditorInfo *')
            self._control(p.ffic.ECTL_GETINFO, info)
            self._info = info
        return info

    @property
    def file_name(self):
        """
        Get file name that is currently being edited.
        """
        p = self._plugin
        size = self._control(p.ffic.ECTL_GETFILENAME, p.ffi.NULL)
        fn = p.ffi.new('wchar_t []', size)
        self._control(p.ffic.ECTL_GETFILENAME, fn)
        return p.f2s(fn)

    def __len__(self):
        """
        Returns number of lines in the editor.
        """
        return self._get_info().TotalLines

    def __getitem__(self, line):
        """
        Returns line text from editor.
        """
        if not 0 <= line < len(self):
            raise IndexError('Line index {} outside of editor line '
                             'range'.format(line))
        p = self._plugin
        egs = p.ffi.new("struct EditorGetString *")
        egs.StringNumber = line
        self._control(p.ffic.ECTL_GETSTRING, egs)
        return self._plugin.f2s(egs.StringText)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    @property
    def size(self):
        """
        Returns editor window size as 2-tuple of (width, height)
        """
        info = self._get_info()
        return (info.WindowSizeX, info.WindowSizeY)

    def _set_position(self, **kwargs):
        self._info = None  # drop cache
        p = self._plugin
        esp = p.ffi.new("struct EditorSetPosition *")
        esp.CurPos = esp.Overtype = esp.CurTabPos\
            = esp.TopScreenLine = esp.LeftPos = -1
        for k, v in kwargs.items():
            setattr(esp, k, v)
        self._control(p.ffic.ECTL_SETPOSITION, esp)

    @property
    def cursor(self):
        """
        Returns cursor position as a 2-tuple of (line, pos)
        """
        info = self._get_info()
        return (info.CurPos, info.CurLine)

    @cursor.setter
    def cursor(self, new_pos):
        x, y = new_pos
        if y > len(self):
            raise IndexError('Setting cursor line to {} is beyond '
                             'last line'.format(y))
        self._set_position(CurLine=y, CurPos=x)

    def focus_cursor(self, x, y):
        """
        Better cursor jumping which takes into account scroll
        to focus on given position better.
        """
        sx, sy = self.scroll
        new_sy = sy
        w, h = self.size
        if y <= sy - 1 or y >= sy + h - 2:
            new_sy = y - h//2
        self.cursor = (x, y)
        if new_sy != sy:
            self.scroll = (sx, new_sy)

    @property
    def scroll(self):
        """
        Returns editor scroll position as 2-tuple in form of
        (left offset, top offset)
        """
        info = self._get_info()
        return (info.LeftPos, info.TopScreenLine)

    @scroll.setter
    def scroll(self, new_pos):
        """
        Changes scroll position. New position is provided as 2-tuple of:
        (left offset, top offset)
        """
        x, y = new_pos
        if y > len(self):
            raise IndexError('Setting scroll line to {} is beyond '
                             'last line'.format(y))
        cx, cy = self.cursor
        # cursor needs to be set when setting scroll, otherwise it
        # will be buggy
        self._set_position(LeftPos=x, TopScreenLine=y, CurPos=cx, CurLine=cy)


class FarPlugin(PluginBase):
    """
    More simplistic FAR plugin interface. Provides functions that
    hide away lower level Plugin C API and allow to write "pythonic" plugins.
    """
    def __init__(self, *args, **kwargs):
        PluginBase.__init__(self, *args, **kwargs)

    def get_panel(self, is_active=True):
        """
        Returns active or passive panel.
        """
        return _Panel(self, is_active)

    def notice(self, body, title=None):
        """
        Simple message box.
        ``body`` is a string or a sequence of strings.
        """
        return self._popup(body, title, 0)

    def error(self, body, title=None):
        """
        Same as .notice(), but message box type is error
        """
        return self._popup(body, title, self.ffic.FMSG_WARNING)

    def _popup(self, body, title, flags):
        items = [self.s2f(title)] if title else [self.s2f('')]
        if isinstance(body, str):
            if '\n' in body:
                body = body.split('\n')
            else:
                items.append(self.s2f(body))
        if not isinstance(body, str):
            items.extend(self.s2f(s) for s in body)
        # note: needs to be at least 2 items, otherwise message 
        # box is not shown
        citems = self.ffi.new('wchar_t *[]', items)
        if not flags:
            flags = self.ffic.FMSG_MB_OK
        return self.info.Message(
            self.info.ModuleNumber,  # GUID
            flags,  # Flags
            self.ffi.NULL,  # HelpTopic
            citems, len(citems), 1,  # ButtonsNumber
        )

    def editor(self, fn, title=None, line=1, column=1, encoding=65001):
        """
        Invokes Far editor on file ``fn``. If ``title`` is given, that
        text is used as editor title, otherwise edited file name is used.

        ``line`` and ``column`` indicate where cursor position will be
        placed after opening.

        Returns True if file was changed, False if it was not and None
        in case of error.
        """
        if title is None:
            title = fn
        fn = self.s2f(fn)
        title = self.s2f(title)
        result = self.info.Editor(fn, title, 0, 0, -1, -1, 
                                  self.ffic.EF_DISABLEHISTORY, line, column, 
                                  encoding)
        if result == self.ffic.EEC_MODIFIED:
            return True
        elif result == self.ffic.EEC_NOT_MODIFIED:
            return False
        return None

    def get_editor(self):
        """
        Get object to control currently opened editor.
        """
        return _Editor(self)

    def menu(self, names, title='', selected=0):
        """
        Simple menu. ``names`` is a list of items. Optional
        ``title`` can be provided.
        """
        items = self.ffi.new('struct FarMenuItem []', len(names))
        refs = []
        for i, name in enumerate(names):
            item = items[i]
            item.Checked = item.Separator = 0
            item.Selected = i == selected
            item.Text = txt = self.s2f(name)
            refs.append(txt)
        title = self.s2f(title)
        NULL = self.ffi.NULL
        return self.info.Menu(self.info.ModuleNumber, -1, -1, 0, 
                              self.ffic.FMENU_AUTOHIGHLIGHT
                              | self.ffic.FMENU_WRAPMODE, title,
                              NULL, NULL, NULL, NULL, items, len(items));




import os
import stat
import time
import logging

from far2l.plugin import PluginVFS
from far2l.farprogress import ProgressMessage
from far2l.fardialogbuilder import (
    Spacer,
    TEXT,
    EDIT,
    PASSWORD,
    MASKED,
    MEMOEDIT,
    BUTTON,
    CHECKBOX,
    RADIOBUTTON,
    COMBOBOX,
    LISTBOX,
    USERCONTROL,
    HLine,
    HSizer,
    VSizer,
    DialogBuilder,
)

from udockerio import Docker

log = logging.getLogger(__name__)

try:
    # pip install runlike
    from runlike import inspector
    class Inspector(inspector.Inspector):
        def inspect(self):
            try:
                output = inspector.check_output(
                    ["docker", "container", "inspect", self.container],
                    stderr=inspector.STDOUT)
                self.container_facts = inspector.loads(output.decode('utf8', 'strict'))
                image_hash = self.get_container_fact("Image")
                output = inspector.check_output(
                    ["docker", "image", "inspect", image_hash],
                    stderr=inspector.STDOUT)
                self.image_facts = inspector.loads(output.decode('utf8', 'strict'))
            except inspector.CalledProcessError as e:
                return str(e)
            return None
except:
    Inspector = None


class DockerBase:

    label = "Python Docker"

    def __init__(self, parent):
        self.parent = parent
        self.info = parent.info
        self.hplugin = parent.hplugin
        self.ffi = parent.ffi
        self.ffic = parent.ffic
        self.s2f = parent.s2f
        self.f2s = parent.f2s
        self.clt = parent.clt

    def Close(self):
        del self.parent
        del self.info
        del self.hplugin
        del self.ffi
        del self.ffic
        del self.s2f
        del self.f2s
        del self.clt

    def OpenPlugin(self, OpenFrom):
        self.names = []
        self.Items = []
        return True

    def Message(self, body, title="Docker", flags=0):
        items = []
        if title:
            items.extend([
                title,
            ])
        items.extend([s for s in body.split('\n')])
        items = [self.s2f(s) for s in items]
        MsgItems = self.ffi.new("wchar_t *[]", items)
        rc = self.info.Message(
            self.info.ModuleNumber,  # GUID
            flags,  # Flags
            self.ffi.NULL,  # HelpTopic
            MsgItems,  # Items
            len(MsgItems),  # ItemsNumber
            1,  # ButtonsNumber
        )
        return rc

    def setName(self, i, name, attr, size):
        self.names[i] = self.s2f(name)
        item = self.Items[i].FindData
        item.dwFileAttributes = attr
        item.nFileSize = size
        item.lpwszFileName = self.names[i]
        return item

    def addName(self, name, attr, size):
        # log.debug('addName({})'.format(name))
        # increase C array
        items = self.ffi.new("struct PluginPanelItem []", len(self.Items) + 1)
        for i in range(len(self.Items)):
            items[i] = self.Items[i]
        self.Items = items
        self.names.append(None)
        self.setName(len(self.Items) - 1, name, attr, size)

    def deleteNames(self, names):
        found = []
        for i in range(len(self.names)):
            if self.f2s(self.names[i]) in names:
                found.append(i)
        if len(found) == 0:
            return False
        # new array
        n = len(self.Items) - len(found)
        if n > 0:
            items = self.ffi.new("struct PluginPanelItem []", n)
            j = 0
            for i in range(len(self.Items)):
                if i not in found:
                    items[j] = self.Items[i]
                    j += 1
            # delete
            for i in sorted(found, reverse=True):
                del self.names[i]
        else:
            self.names = []
            items = []
        self.Items = items
        #self.info.Control(self.hplugin, self.ffic.FCTL_UPDATEPANEL, 0, 0)
        #self.info.Control(self.hplugin, self.ffic.FCTL_REDRAWPANEL, 0, 0)
        return True

    def doRefresh(self):
        self.info.Control(self.hplugin, self.ffic.FCTL_UPDATEPANEL, 0, 0)
        self.info.Control(self.hplugin, self.ffic.FCTL_REDRAWPANEL, 0, 0)


class DockerDevice(DockerBase):
    def __init__(self, parent, device):
        super().__init__(parent)
        self.device = device
        self.devicepath = "/"

    def addResult(self, result):
        n = 0
        for rec in result:
            if rec.st_name in (".", ".."):
                continue
            n += 1
        self.Items = self.ffi.new("struct PluginPanelItem []", n)
        self.names = [None] * n
        i = 0
        for rec in result:
            if rec.st_name in (".", ".."):
                continue
            attr = 0
            attr |= self.ffic.FILE_ATTRIBUTE_DIRECTORY if rec.st_mode & stat.S_IFDIR else 0
            attr |= self.ffic.FILE_ATTRIBUTE_DEVICE if rec.st_mode & stat.S_IFCHR else 0
            attr |= self.ffic.FILE_ATTRIBUTE_ARCHIVE if rec.st_mode & stat.S_IFREG else 0
            # log.debug('{} mode={:5o} attr={} perms={}'.format(rec.name, rec.mode, attr, rec.perms))
            # datetime.datetime.fromtimestamp(rec.time).strftime('%Y-%m-%d %H:%M:%S')
            item = self.setName(i, rec.st_name, attr, rec.st_size)
            item.dwUnixMode = rec.st_mode

            t = (
                int(time.mktime(rec.st_mtime.timetuple())) + self.ffic.EPOCH_DIFFERENCE
            ) * self.ffic.TICKS_PER_SECOND
            item.ftLastWriteTime.dwHighDateTime = t >> 32
            item.ftLastWriteTime.dwLowDateTime = t & 0xFFFFFFFF
            i += 1

    def loadDirectory(self):
        result = self.clt.ls(self.device, self.devicepath)
        self.addResult(result)

    def devGetFile(self, sqname, dqname):
        self.clt.pull(self.device, sqname, dqname)

    def devPutFile(self, sqname, dqname):
        self.clt.push(self.device, sqname, dqname)

    def devDelete(self, dqname):
        self.clt.remove(self.device, dqname)

    def devMakeDirectory(self, dqname):
        self.clt.mkdir(self.device, dqname)

    def GetOpenPluginInfo(self, OpenInfo):
        #
        # WARNING WARNING - dangerous pointers and lifetime of a variable in python
        # anything passed as a pointer must be kept in local python variables
        # otherwise gc/cffi will delete them and miracles will happen
        #
        self.title = f"docker://{self.device}{self.devicepath}"
        self._curdir = self.s2f(self.devicepath)
        self._title = self.s2f(self.title)
        self._label = self.s2f(self.label)

        Info = self.ffi.cast("struct OpenPluginInfo *", OpenInfo)
        Info.Flags = (
            self.ffic.OPIF_USEFILTER
            | self.ffic.OPIF_USESORTGROUPS
            | self.ffic.OPIF_USEHIGHLIGHTING
            | self.ffic.OPIF_ADDDOTS
            | self.ffic.OPIF_SHOWNAMESONLY
        )
        Info.HostFile = self.ffi.NULL
        Info.CurDir = self._curdir
        Info.Format = self._label
        Info.PanelTitle = self._title
        # const struct InfoPanelLine *InfoLines;
        # int                   InfoLinesNumber;
        # const wchar_t * const   *DescrFiles;
        # int                   DescrFilesNumber;
        # const struct PanelMode *PanelModesArray;
        Info.PanelModesNumber = 0
        Info.StartPanelMode = 0
        # Info.StartSortMode = self.ffic.SM_NAME
        # Info.StartSortOrder = 0
        # const struct KeyBarTitles *KeyBar;
        # Info.ShortcutData = self.s2f('py:adb cd '+title)

    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        try:
            self.loadDirectory()
        except Exception as ex:
            log.exception("unknown exception:")
            self.Message(str(ex), "Unknown exception.", self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
            return False
        return True

    def SetDirectory(self, Dir, OpMode):
        name = self.f2s(Dir)
        if name == "..":
            if self.devicepath == "/":
                self.parent.handler = DockerDevices(self.parent)
                self.Close()
            else:
                self.devicepath = os.path.normpath(os.path.join(self.devicepath, name))
        else:
            self.devicepath = os.path.join(self.devicepath, name)
        return True

    def GetFiles(self, PanelItem, ItemsNumber, Move, DestPath, OpMode):
        items = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        name = self.f2s(items[ItemsNumber-1].FindData.lpwszFileName)
        t = ProgressMessage(self, "Copying", "Please wait ... working", 100)
        t.show()
        DestPath = self.ffi.cast("wchar_t **", DestPath)
        dpath = self.ffi.string(DestPath[0])
        # log.debug('GetFiles: {} {} OpMode={}'.format(ItemsNumber, OpMode, dpath))
        for i in range(ItemsNumber):
            t.update(i)
            name = self.f2s(items[i].FindData.lpwszFileName)
            if name in (".", ".."):
                continue
            sqname = os.path.join(self.devicepath, name)
            dqname = os.path.join(dpath, name)
            # log.debug('pull: {} -> {} OpMode={}'.format(sqname, dqname, OpMode))
            try:
                self.devGetFile(sqname, dqname)
            except Exception as ex:
                t.close()
                log.exception("unknown exception:")
                self.Message(str(ex), "GetFile exception.", self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
                return False
        t.close()
        return True

    def PutFiles(self, PanelItem, ItemsNumber, Move, SrcPath, OpMode):
        items = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        spath = self.f2s(SrcPath)
        for i in range(ItemsNumber):
            name = self.f2s(items[i].FindData.lpwszFileName)
            if name in (".", ".."):
                continue
            sqname = os.path.join(spath, name)
            dqname = os.path.join(self.devicepath, name)
            # log.debug('push: {} -> {} OpMode={}'.format(sqname, dqname, OpMode))
            try:
                self.devPutFile(sqname, dqname)
            except Exception as ex:
                log.exception("unknown exception:")
                self.Message(str(ex), "PutFile exception.", self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
                return False
        return True

    def DeleteFiles(self, PanelItem, ItemsNumber, OpMode):
        items = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        if ItemsNumber > 1:
            rc = self.Message("Do you wish to delete the files ?", flags=self.ffic.FMSG_MB_YESNO)
        else:
            name = self.f2s(items[0].FindData.lpwszFileName)
            rc = self.Message(f"Do you wish to delete the file:\n{name}", flags=self.ffic.FMSG_MB_YESNO)
        if rc == 1: # NO
            return True
        for i in range(ItemsNumber):
            name = self.f2s(items[i].FindData.lpwszFileName)
            if name in (".", ".."):
                continue
            dqname = os.path.join(self.devicepath, name)
            log.debug("remove: {}, OpMode={}".format(dqname, OpMode))
            try:
                self.devDelete(dqname)
            except Exception as ex:
                log.exception("unknown exception:")
                self.Message(str(ex), "Unknown exception.", self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
                return False
        return True

    def MakeDirectory(self, Name, OpMode):
        name = self.ffi.cast("wchar_t **", Name)
        name = self.ffi.string(name[0])

        b = DialogBuilder(
            self,
            self.info.DefDlgProc,
            "Make folder",
            "makefolder",
            0,
            VSizer(
                TEXT(None, "Create &folder"),
                EDIT("path", 36, maxlength=40),
                HLine(),
                HSizer(
                    BUTTON("OK", "OK", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("CANCEL", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        if res == dlg.ID_OK:
            path = dlg.GetText(dlg.ID_path).strip()
            dqname = os.path.normpath(os.path.join(self.devicepath, path))
            log.debug("mkdir {}".format(dqname))
            if path:
                try:
                    self.devMakeDirectory(dqname)
                except Exception as ex:
                    log.exception("unknown exception:")
                    self.Message(str(ex), "Unknown exception.", self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
                    return False
        self.info.DialogFree(dlg.hDlg)

        return True

    def EditAttributes(self):
        item, ppidata = self.parent.panel.GetCurrentPanelItem()
        name = self.f2s(item.FindData.lpwszFileName)
        fqname = os.path.normpath(os.path.join(self.devicepath, name))
        self.Message(f"Current item: {fqname}", "TODO EditAttributes", self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)

    def ProcessKey(self, Key, ControlState):
        if (
            Key == self.ffic.KEY_CTRLA - self.ffic.KEY_CTRL
            and ControlState == self.ffic.PKF_CONTROL
        ):
            self.EditAttributes()
            return 1
        return 0


class DockerDevices(DockerBase):
    def __init__(self, parent):
        super().__init__(parent)
        self.devicepath = '/'

    def addResult(self, result):
        self.Items = self.ffi.new("struct PluginPanelItem []", len(result))
        self.names = [None] * len(result)
        i = 0
        for rec in result:
            if rec[2]:
                self.setName(i, rec[1], self.ffic.FILE_ATTRIBUTE_DIRECTORY, 0)
            else:
                self.setName(i, rec[1], self.ffic.FILE_ATTRIBUTE_OFFLINE, 0)
            i += 1

    def devStart(self, name):
        self.clt.start(name)

    def devStop(self, name):
        self.clt.stop(name)

    def devLoadDevices(self):
        result = self.clt.list()
        self.addResult(result)

    def GetOpenPluginInfo(self, OpenInfo):
        #
        # WARNING WARNING - dangerous pointers and lifetime of a variable in python
        # anything passed as a pointer must be kept in local python variables
        # otherwise gc/cffi will delete them and miracles will happen
        #
        self.title = self.label
        self._curdir = self.s2f(self.devicepath)
        self._title = self.s2f(self.title)
        self._label = self.s2f(self.label)

        #self._nokey = self.ffi.cast("wchar_t *", self.ffi.NULL)
        self._nokey = self.s2f("")
        py_keybar_normal = [self._nokey]*12
        py_keybar_ctrl = [self._nokey]*12
        py_keybar_alt = [self._nokey]*12
        py_keybar_ctrl_shift = [self._nokey]*12
        py_keybar_alt_shift = [self._nokey]*12
        py_keybar_ctrl_alt = [self._nokey]*12

        py_keybar_normal[2-1]=self.s2f("Ena/Dis")
        py_keybar_normal[3-1]=self.s2f("Run Like")

        self.py_keybar = [
            py_keybar_normal,
            py_keybar_ctrl,
            py_keybar_alt,
            py_keybar_ctrl_shift,
            py_keybar_alt_shift,
            py_keybar_ctrl_alt
        ]
        self._keybar = self.ffi.new("struct KeyBarTitles *", self.py_keybar)

        Info = self.ffi.cast("struct OpenPluginInfo *", OpenInfo)
        Info.Flags = (
            self.ffic.OPIF_USEFILTER
            | self.ffic.OPIF_USESORTGROUPS
            | self.ffic.OPIF_USEHIGHLIGHTING
            | self.ffic.OPIF_ADDDOTS
            | self.ffic.OPIF_SHOWNAMESONLY
        )
        Info.HostFile = self.ffi.NULL
        Info.CurDir = self._curdir
        Info.Format = self._label
        Info.PanelTitle = self._title
        # const struct InfoPanelLine *InfoLines;
        # int                   InfoLinesNumber;
        # const wchar_t * const   *DescrFiles;
        # int                   DescrFilesNumber;
        # const struct PanelMode *PanelModesArray;
        Info.PanelModesNumber = 0
        Info.StartPanelMode = 0
        # Info.StartSortMode = self.ffic.SM_NAME
        # Info.StartSortOrder = 0
        Info.KeyBar = self._keybar
        # Info.ShortcutData = self.s2f('py:adb cd '+title)

    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        try:
            self.devLoadDevices()
        except Exception as ex:
            log.exception("unknown exception:")
            self.Message(str(ex), "Unknown exception.", self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
            return False
        return True

    def SetDirectory(self, Dir, OpMode):
        name = self.f2s(Dir)
        if name == "":
            self.info.Control(self.hplugin, self.ffic.FCTL_CLOSEPLUGIN, 0, 0)
            return True
        if name == "..":
            self.info.Control(self.hplugin, self.ffic.FCTL_CLOSEPLUGIN, 0, Dir)
            return True
        self.parent.handler = DockerDevice(self.parent, name)
        self.Close()
        return True

    def GetFiles(self, PanelItem, ItemsNumber, Move, DestPath, OpMode):
        self.Message("GetFiles allowed only inside device.", flags=self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
        return False

    def PutFiles(self, PanelItem, ItemsNumber, Move, SrcPath, OpMode):
        self.Message("PutFiles allowed only inside device.", flags=self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
        return False

    def DeleteFiles(self, PanelItem, ItemsNumber, OpMode):
        self.Message("DeleteFiles allowed only inside device.", flags=self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
        return False

    def MakeDirectory(self, Name, OpMode):
        self.Message("MakeDirectory allowed only inside device.", flags=self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
        return False

    def CMDStartStop(self, item, name):
        running = (item.FindData.dwFileAttributes & self.ffic.FILE_ATTRIBUTE_DIRECTORY) != 0
        proc = self.devStop if running else self.devStart
        try:
            proc(name)
            self.doRefresh()
        except Exception as ex:
            log.exception("unknown exception:")
            self.Message(str(ex), "Unknown exception.", self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)

    def CMDRunLike(self, item, name):
        if Inspector is None:
            self.Message("Package runlike required but not installed.", flags=self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
            return
        running = (item.FindData.dwFileAttributes & self.ffic.FILE_ATTRIBUTE_DIRECTORY) != 0
        insp = Inspector(name, None, True)
        error = insp.inspect()
        if error is not None:
            self.Message(error, "RunLike error.", flags=self.ffic.FMSG_MB_OK|self.ffic.FMSG_WARNING)
            return
        open("/tmp/uinfo", "wt").write(f'''\
container name={name} running={running}

{insp.format_cli()}
''')
        self.info.Editor(
            "/tmp/uinfo",
            "uinfo",
            0, 0, -1, -1,
            #0,0,50,25,
            self.ffic.EF_DISABLEHISTORY,#|self.ffic.EF_NONMODAL|self.ffic.EF_IMMEDIATERETURN
            0,
            0,
            0xFFFFFFFF,  # =-1=self.ffic.CP_AUTODETECT
        )
        os.unlink('/tmp/uinfo')

    def ProcessKey(self, Key, ControlState):
        if Key == self.ffic.VK_F2 and ControlState == 0:
            item, ppidata = self.parent.panel.GetCurrentPanelItem()
            name = self.f2s(item.FindData.lpwszFileName)
            if name == "..":
                return
            self.CMDStartStop(item, name)
            return 1
        if Key == self.ffic.VK_F3 and ControlState == 0:
            item, ppidata = self.parent.panel.GetCurrentPanelItem()
            name = self.f2s(item.FindData.lpwszFileName)
            if name == "..":
                return
            self.CMDRunLike(item, name)
            return 1
        if Key == self.ffic.KEY_ENTER and ControlState == self.ffic.PKF_CONTROL:
            log.debug("ProcessKey: CTRL+ENTER")
            return 1
        return 0


class Plugin(PluginVFS):
    label = "Python Docker"
    openFrom = ["PLUGINSMENU", "DISKMENU"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clt = Docker()
        self.handler = DockerDevices(self)

    def OpenPlugin(self, OpenFrom):
        return self.handler.OpenPlugin(OpenFrom)

    def GetOpenPluginInfo(self, OpenInfo):
        return self.handler.GetOpenPluginInfo(OpenInfo)

    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        if OpMode & self.ffic.OPM_FIND:
            # find in root of device causes core dump on linux
            return False
        ok = self.handler.GetFindData(PanelItem, ItemsNumber, OpMode)
        if ok:
            PanelItem = self.ffi.cast("struct PluginPanelItem **", PanelItem)
            ItemsNumber = self.ffi.cast("int *", ItemsNumber)
            n = len(self.handler.Items)
            p = self.ffi.NULL if n == 0 else self.handler.Items
            PanelItem[0] = p
            ItemsNumber[0] = n
        return ok

    def FreeFindData(self, PanelItem, ItemsNumber):
        log.debug(f"FreeFindData({PanelItem}, {ItemsNumber}, n.Items={len(self.handler.Items)})")
        self.handler.names = []
        self.handler.Items = []

    def SetDirectory(self, Dir, OpMode):
        return self.handler.SetDirectory(Dir, OpMode)

    def Compare(self, PanelItem1, PanelItem2, Mode):
        def cmp(a, b):
            if a < b:
                return -1
            elif a == b:
                return 0
            return 1

        p1 = self.ffi.cast("struct PluginPanelItem *", PanelItem1)
        p2 = self.ffi.cast("struct PluginPanelItem *", PanelItem2)
        n1 = self.f2s(p1.FindData.lpwszFileName)
        n2 = self.f2s(p2.FindData.lpwszFileName)
        rc = cmp(n1, n2)
        # log.debug('Compare: cmp({}, {})={}, mode={}'.format(n1, n2, rc, Mode))
        return rc

    def GetVirtualFindData(self, PanelItem, ItemsNumber, Path):
        return False

    def FreeVirtualFindData(self, PanelItem, ItemsNumber):
        log.debug("FreeVirtualFindData: {}, {}".format(PanelItem, ItemsNumber))

    def GetFiles(self, PanelItem, ItemsNumber, Move, DestPath, OpMode):
        if ItemsNumber == 0 or Move:
            return False
        return self.handler.GetFiles(PanelItem, ItemsNumber, Move, DestPath, OpMode)

    def PutFiles(self, PanelItem, ItemsNumber, Move, SrcPath, OpMode):
        if ItemsNumber == 0 or Move:
            return False
        return self.handler.PutFiles(PanelItem, ItemsNumber, Move, SrcPath, OpMode)

    def DeleteFiles(self, PanelItem, ItemsNumber, OpMode):
        if ItemsNumber == 0:
            return False
        return self.handler.DeleteFiles(PanelItem, ItemsNumber, OpMode)

    def MakeDirectory(self, Name, OpMode):
        return self.handler.MakeDirectory(Name, OpMode)

    def ProcessEvent(self, Event, Param):
        return 0

    def ProcessKey(self, Key, ControlState):
        return self.handler.ProcessKey(Key, ControlState)
